 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of getting the base message registry
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

def get_base_registry(_redfishobj):

    registries_uri = _redfishobj.root.obj['Registries']['@odata.id']

    if registries_uri:
        registries_members = _redfishobj.get(registries_uri).obj['Members']
        for registry in registries_members:
            sys.stdout.write("Registry URI at '%s'\n" % registry['@odata.id'])

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    get_base_registry(REDFISHOBJ)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
"""A threaded version of RIS _load for quicker searching"""
# ---------Imports---------

import logging
import threading
from queue import Empty

import jsonpath_rw

# Added for py3 compatibility
import six
from six.moves.urllib.parse import urlparse, urlunparse

import redfish.ris

# ---------End of imports---------

# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)

# ---------End of debug logger---------


class LoadWorker(threading.Thread):
    """A threaded implementation of _load for quicker crawling"""

    def __init__(self, queue):
        threading.Thread.__init__(self)
        self.queue = queue
        self.exception = None

    def run(self):
        """Main worker function"""
        try:
            while True:
                (
                    path,
                    includelogs,
                    loadcomplete,
                    crawl,
                    rel,
                    init,
                    prevpath,
                    originaluri,
                    theobj,
                ) = self.queue.get()
                if (
                    path
                    == includelogs
                    == loadcomplete
                    == crawl
                    == rel
                    == init
                    == prevpath
                    == originaluri
                    == theobj
                    == "KILL"
                ):
                    break

                if path.endswith("?page=1") and not loadcomplete:
                    # Don't download schemas in crawl unless we are loading absolutely everything
                    self.queue.task_done()
                    continue
                elif not includelogs and crawl:
                    # Only include logs when asked as there can be an extreme amount of entries
                    if "/log" in path.lower():
                        self.queue.task_done()
                        continue

                # TODO: need to find a better way to support non ascii characters
                path = path.replace("|", "%7C")
                # remove fragments
                newpath = urlparse(path)
                newpath = list(newpath[:])
                newpath[-1] = ""
                path = urlunparse(tuple(newpath))

                if prevpath and prevpath != path:
                    theobj.ctree[prevpath].update([path])
                if not rel:
                    if path.lower() in theobj.visited_urls:
                        self.queue.task_done()
                        continue
                LOGGER.debug("_loading %s", path)

                resp = theobj._client.get(path)

                if resp.status != 200 and path.lower() == theobj.typepath.defs.biospath:
                    self.queue.task_done()
                    raise redfish.ris.ris.BiosUnregisteredError()
                elif resp.status == 401:
                    self.queue.task_done()
                    raise redfish.ris.ris.SessionExpired(
                        "Invalid session. Please logout and " "log back in or include credentials."
                    )
                elif resp.status not in (201, 200):
                    theobj.removepath(path)
                    self.queue.task_done()
                    continue

                theobj.update_member(resp=resp, path=path, init=init)

                fpath = (
                    lambda pa, path: path
                    if pa.endswith(theobj.typepath.defs.hrefstring) and pa.startswith((theobj.collstr, "Entries"))
                    else None
                )

                # follow all the href attributes
                if theobj.is_redfish:
                    jsonpath_expr = jsonpath_rw.parse("$..'@odata.id'")
                else:
                    jsonpath_expr = jsonpath_rw.parse("$..href")
                matches = jsonpath_expr.find(resp.dict)

                if "links" in resp.dict and "NextPage" in resp.dict["links"]:
                    if originaluri:
                        next_link_uri = originaluri + "?page=" + str(resp.dict["links"]["NextPage"]["page"])
                        href = "%s" % next_link_uri

                        theobj.get_queue.put(
                            (
                                href,
                                includelogs,
                                loadcomplete,
                                crawl,
                                rel,
                                init,
                                None,
                                originaluri,
                                theobj,
                            )
                        )
                    else:
                        next_link_uri = path + "?page=" + str(resp.dict["links"]["NextPage"]["page"])

                        href = "%s" % next_link_uri
                        theobj.get_queue.put(
                            (
                                href,
                                includelogs,
                                loadcomplete,
                                crawl,
                                rel,
                                init,
                                None,
                                path,
                                theobj,
                            )
                        )

                # Only use monolith if we are set to
                matchrdirpath = (
                    next(
                        (match for match in matches if match.value == theobj._resourcedir),
                        None,
                    )
                    if theobj.directory_load
                    else None
                )
                if not matchrdirpath and crawl:
                    for match in matches:
                        if path == "/rest/v1" and not loadcomplete:
                            if (
                                str(match.full_path) == "links.Schemas.href"
                                or str(match.full_path) == "links.Registries.href"
                            ):
                                continue
                        elif not loadcomplete:
                            if (
                                str(match.full_path) == "Registries.@odata.id"
                                or str(match.full_path) == "JsonSchemas.@odata.id"
                            ):
                                continue

                        if match.value == path:
                            continue
                        elif not isinstance(match.value, six.string_types):
                            continue

                        href = "%s" % match.value
                        theobj.get_queue.put(
                            (
                                href,
                                includelogs,
                                loadcomplete,
                                crawl,
                                rel,
                                init,
                                fpath(str(match.full_path), path),
                                originaluri,
                                theobj,
                            )
                        )
                elif crawl:
                    href = "%s" % matchrdirpath.value
                    theobj.get_queue.put(
                        (
                            href,
                            includelogs,
                            loadcomplete,
                            crawl,
                            rel,
                            init,
                            path,
                            originaluri,
                            theobj,
                        )
                    )
                if loadcomplete:
                    if path == "/rest/v1":
                        schemamatch = jsonpath_rw.parse("$..extref")
                    else:
                        schemamatch = jsonpath_rw.parse("$..Uri")
                    smatches = schemamatch.find(resp.dict)
                    matches = matches + smatches
                    for match in matches:
                        if isinstance(match.value, six.string_types):
                            theobj.get_queue.put(
                                (
                                    match.value,
                                    includelogs,
                                    loadcomplete,
                                    crawl,
                                    rel,
                                    init,
                                    fpath(str(match.full_path), path),
                                    originaluri,
                                    theobj,
                                )
                            )
                self.queue.task_done()
        except Empty:
            pass
        except Exception as excp:
            self.exception = excp

    def get_exception(self):
        """Get any exception from the thread"""
        return self.exception



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with blob store interface"""

# ---------Imports---------

import logging
import os
import random
import string
import struct
import sys
import time
import os
from ctypes import (
    POINTER,
    c_char_p,
    c_ubyte,
    c_uint,
    c_ushort,
    c_void_p,
    cdll,
    create_string_buffer,
)

from redfish.hpilo.rishpilo import BlobReturnCodes as hpiloreturncodes
from redfish.hpilo.rishpilo import (
    HpIlo,
    HpIloChifPacketExchangeError,
    HpIloInitialError,
)

if os.name == "nt":
    from ctypes import windll
else:
    from _ctypes import dlclose

# ---------End of imports---------
# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------
# -----------------------Error Returns----------------------


class UnexpectedResponseError(Exception):
    """Raise when we get data that we don't expect from iLO"""

    pass


class HpIloError(Exception):
    """Raised when iLO returns non-zero error code"""

    pass


class Blob2CreateError(Exception):
    """Raised when create operation fails"""

    pass


class Blob2InfoError(Exception):
    """Raised when create operation fails"""

    pass


class Blob2ReadError(Exception):
    """Raised when read operation fails"""

    pass


class Blob2WriteError(Exception):
    """Raised when write operation fails"""

    pass


class Blob2DeleteError(Exception):
    """Raised when delete operation fails"""

    pass


class Blob2OverrideError(Exception):
    """Raised when delete operation fails because of it been overwritten"""

    pass


class BlobRetriesExhaustedError(Exception):
    """Raised when max retries have been attempted for same operation"""

    pass


class Blob2FinalizeError(Exception):
    """Raised when finalize operation fails"""

    pass


class Blob2ListError(Exception):
    """Raised when list operation fails"""

    pass


class Blob2SecurityError(Exception):
    """Raised when there is an issue with security"""

    pass


class BlobNotFoundError(Exception):
    """Raised when blob not found in key/namespace"""

    pass


class ChifDllMissingError(Exception):
    """Raised when unable to obtain ilorest_chif dll handle"""

    pass


class EncryptionEnabledError(Exception):
    """Raised when high security encryption is enabled"""

    pass


# ----------------------------------------------------------

# -------------------Helper functions-------------------------


class BlobReturnCodes(object):
    """Blob store return codes.

    SUCCESS           success
    BADPARAMETER      bad parameter supplied
    NOTFOUND          blob name not found
    NOTMODIFIED       call did not perform the operation

    """

    SUCCESS = 0
    BADPARAMETER = 2
    NOTFOUND = 12
    NOTMODIFIED = 20


class BlobStore2(object):
    """Blob store 2 class"""

    def __init__(self, log_dir=None):
        lib = self.gethprestchifhandle()
        self.log_dir = log_dir
        self.channel = HpIlo(dll=lib, log_dir=log_dir)
        self.max_retries = 40
        self.max_read_retries = 3
        self.delay = 0.25

    def __del__(self):
        """Blob store 2 close channel function"""
        if hasattr(self, "channel"):
            self.channel.close()

    def create(self, key, namespace):
        """Create the blob

        :param key: The blob key to create.
        :type key: str.
        :param namespace: The blob namespace to create the key in.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.create_not_blobentry.argtypes = [c_char_p, c_char_p]
        lib.create_not_blobentry.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.create_not_blobentry(name, namespace)
        data = ptr[: lib.size_of_createRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def get_info(self, key, namespace, retries=0):
        """Get information for a particular blob

        :param key: The blob key to retrieve.
        :type key: str.
        :param namespace: The blob namespace to retrieve the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.get_info.argtypes = [c_char_p, c_char_p]
        lib.get_info.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namspace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.get_info(name, namspace)
        data = ptr[: lib.size_of_infoRequest()]
        data = bytearray(data)

        while(retries <= self.max_retries):
            """The Redfish request is submitted to iLO successfully in rest_immediate() function.
            Here we are checking if iLO has processed the Redfish request and the response is 
            written in the key in the volatile namespace in blob. 
            If iLO has not processed the request and response is not written, we get BADPARAMETER error. 
            Hence retry after 0.25 seconds."""
            resp = self._send_receive_raw(data)

            #checking for errors in the received response
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]

            if errorcode == BlobReturnCodes.BADPARAMETER:
                #if BADPARAMETER, delaying the execution by 0.25 seconds and retrying "send_receive_raw"
                if retries < self.max_retries:
                    time.sleep(self.delay)
                    retries += 1
                else:
                    #if all retries are exhausted and an error is still received, raise Blob2override error.
                    raise Blob2OverrideError(errorcode)
            elif errorcode == BlobReturnCodes.NOTFOUND:
                raise BlobNotFoundError(key, namespace)
            elif not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
                raise HpIloError(errorcode)
            else:
                break

        response = resp[lib.size_of_responseHeaderBlob() :]

        self.unloadchifhandle(lib)

        return response

    def read(self, key, namespace, retries=0):
        """Read a particular blob

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        maxread = lib.max_read_size()
        readsize = lib.size_of_readRequest()
        readhead = lib.size_of_responseHeaderBlob()

        self.unloadchifhandle(lib)

        blob_info = self.get_info(key, namespace)
        blobsize = struct.unpack("<I", bytes(blob_info[0:4]))[0]

        bytes_read = 0
        data = bytearray()

        while bytes_read < blobsize:
            if (maxread - readsize) < (blobsize - bytes_read):
                count = maxread - readsize
            else:
                count = blobsize - bytes_read

            read_block_size = bytes_read
            recvpkt = self.read_fragment(key, namespace, read_block_size, count)

            newreadsize = readhead + 4
            bytesread = struct.unpack("<I", bytes(recvpkt[readhead:(newreadsize)]))[0]

            if bytesread == 0:
                if retries < self.max_read_retries:
                    data = self.read(key=key, namespace=namespace, retries=retries + 1)
                    return data
                else:
                    raise BlobRetriesExhaustedError()

            data.extend(recvpkt[newreadsize : newreadsize + bytesread])
            bytes_read += bytesread

        return data

    def read_fragment(self, key, namespace, offset=0, count=1):
        """Fragmented version of read function for large blobs

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.
        :param offset: The data offset for the current fragmented read.
        :type key: int.
        :param count: The data count for the current fragmented read.
        :type namespace: int.

        """
        lib = self.gethprestchifhandle()
        lib.read_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
        lib.read_fragment.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.read_fragment(offset, count, name, namespace)
        data = ptr[: lib.size_of_readRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        resp = resp + b"\0" * (lib.size_of_readResponse() - len(resp))

        return resp

    def write(self, key, namespace, data=None):
        """Write a particular blob

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written.
        :type data: str.

        """
        lib = self.gethprestchifhandle()
        maxwrite = lib.max_write_size()
        writesize = lib.size_of_writeRequest()

        self.unloadchifhandle(lib)

        if data:
            data_length = len(data)
            bytes_written = 0

            while bytes_written < data_length:
                if (maxwrite - writesize) < (data_length - bytes_written):
                    count = maxwrite - writesize
                else:
                    count = data_length - bytes_written

                write_blob_size = bytes_written

                self.write_fragment(
                    key,
                    namespace=namespace,
                    data=data[write_blob_size : write_blob_size + count],
                    offset=write_blob_size,
                    count=count,
                )

                bytes_written += count

        return self.finalize(key, namespace=namespace)

    def write_fragment(self, key, namespace, data=None, offset=0, count=1):
        """Fragmented version of write function for large blobs

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written to blob.
        :type data: str.
        :param offset: The data offset for the current fragmented write.
        :type key: int.
        :param count: The data count for the current fragmented write.
        :type count: int.

        """
        lib = self.gethprestchifhandle()
        lib.write_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
        lib.write_fragment.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.write_fragment(offset, count, name, namespace)
        sendpacket = ptr[: lib.size_of_writeRequest()]

        if isinstance(data, str):
            data = data.encode("utf-8")

        dataarr = bytearray(sendpacket)
        dataarr.extend(memoryview(data))

        resp = self._send_receive_raw(dataarr)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def delete(self, key, namespace, retries=0):
        """Delete the blob

        :param key: The blob key to be deleted.
        :type key: str.
        :param namespace: The blob namespace to delete the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.delete_blob.argtypes = [c_char_p, c_char_p]
        lib.delete_blob.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namspace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.delete_blob(name, namspace)
        data = ptr[: lib.size_of_deleteRequest()]
        data = bytearray(data)

        while(retries <= self.max_retries):
            """The Redfish request is submitted to iLO successfully in rest_immediate() function.
            Here we are checking if iLO has processed the Redfish request and the response is 
            written in the key in the volatile namespace in blob. 
            If iLO has not processed the request and response is not written, we get BADPARAMETER error. 
            Hence retry after 0.25 seconds."""
            resp = self._send_receive_raw(data)

            #checking for errors in the received response
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]

            if errorcode == BlobReturnCodes.BADPARAMETER:
                #if error is BADPARAMETER, delaying the execution by 0.25 seconds and retrying "send_receive_raw"
                if retries < self.max_retries:
                    time.sleep(self.delay)
                    retries += 1
                else:
                    #if all retries are exhausted and an error is still received, raise Blob2override error.
                    raise Blob2OverrideError(errorcode)
            elif not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
                raise HpIloError(errorcode)
            else:
                break

        self.unloadchifhandle(lib)

        return errorcode

    def list(self, namespace):
        """List operation to retrieve all blobs in a given namespace

        :param namespace: The blob namespace to retrieve the keys from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.list_blob.argtypes = [c_char_p]
        lib.list_blob.restype = POINTER(c_ubyte)

        namespace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.list_blob(namespace)
        data = ptr[: lib.size_of_listRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        resp = resp + b"\0" * (lib.size_of_listResponse() - len(resp))

        self.unloadchifhandle(lib)

        return resp

    def finalize(self, key, namespace):
        """Finalize the blob

        :param key: The blob key to be finalized.
        :type key: str.
        :param namespace: The blob namespace to finalize the key in.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.finalize_blob.argtypes = [c_char_p, c_char_p]
        lib.finalize_blob.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.finalize_blob(name, namespace)
        data = ptr[: lib.size_of_finalizeRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return errorcode

    def rest_immediate(
        self,
        req_data,
        rqt_key="RisRequest",
        rsp_key="RisResponse",
        rsp_namespace="volatile",
    ):
        """Read/write blob via immediate operation

        :param req_data: The blob data to be read/written.
        :type req_data: str.
        :param rqt_key: The blob key to be used for the request data.
        :type rqt_key: str.
        :param rsp_key: The blob key to be used for the response data.
        :type rsp_key: str.
        :param rsp_namespace: The blob namespace to retrieve the response from.
        :type rsp_namespace: str.

        """
        rqt_key = "".join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        rsp_key = "".join(random.choice(string.ascii_letters + string.digits) for _ in range(10))

        lib = self.gethprestchifhandle()

        if len(req_data) < (lib.size_of_restImmediateRequest() + lib.max_write_size()):
            lib.rest_immediate.argtypes = [c_uint, c_char_p, c_char_p]
            lib.rest_immediate.restype = POINTER(c_ubyte)

            name = create_string_buffer(rsp_key.encode("utf-8"))
            namespace = create_string_buffer(rsp_namespace.encode("utf-8"))

            ptr = lib.rest_immediate(len(req_data), name, namespace)
            sendpacket = ptr[: lib.size_of_restImmediateRequest()]
            mode = False
        else:
            self.create(rqt_key, rsp_namespace)
            self.write(rqt_key, rsp_namespace, req_data)

            lib.rest_immediate_blobdesc.argtypes = [c_char_p, c_char_p, c_char_p]
            lib.rest_immediate_blobdesc.restype = POINTER(c_ubyte)

            name = create_string_buffer(rqt_key.encode("utf-8"))
            namespace = create_string_buffer(rsp_namespace.encode("utf-8"))
            rspname = create_string_buffer(rsp_key.encode("utf-8"))

            ptr = lib.rest_immediate_blobdesc(name, rspname, namespace)
            sendpacket = ptr[: lib.size_of_restBlobRequest()]
            mode = True

        data = bytearray(sendpacket)

        if not mode:
            data.extend(req_data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if errorcode == BlobReturnCodes.NOTFOUND:
            raise BlobNotFoundError(rsp_key, rsp_namespace)

        recvmode = struct.unpack("<I", bytes(resp[12:16]))[0]

        fixdlen = lib.size_of_restResponseFixed()
        response = resp[fixdlen : struct.unpack("<I", bytes(resp[16:20]))[0] + fixdlen]

        tmpresponse = None
        if errorcode == BlobReturnCodes.SUCCESS and not mode:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
        elif errorcode == BlobReturnCodes.NOTMODIFIED and not mode:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
        elif errorcode == BlobReturnCodes.SUCCESS:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
        elif recvmode == 0:
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        if not tmpresponse and recvmode == 1:
            tmpresponse = self.read(rsp_key, rsp_namespace)

            try:
                self.delete(rsp_key, rsp_namespace)
            except Exception as excp:
                raise excp
        else:
            try:
                self.delete(rsp_key, rsp_namespace)
            except Blob2OverrideError:
                pass
            except HpIloChifPacketExchangeError:
                pass
            except Exception as excp:
                raise excp

        return tmpresponse

    def get_security_state(self):
        """Get information for the current security state"""
        lib = self.gethprestchifhandle()
        lib.get_security_state.argtypes = []
        lib.get_security_state.restype = POINTER(c_ubyte)

        ptr = lib.get_security_state()
        data = ptr[: lib.size_of_securityStateRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        try:
            securitystate = struct.unpack("<c", bytes(resp[72]))[0]
        except:
            securitystate = int(resp[72])

        self.unloadchifhandle(lib)

        return securitystate

    def mount_blackbox(self):
        """Operation to mount the blackbox partition"""
        lib = self.gethprestchifhandle()
        lib.blackbox_media_mount.argtypes = []
        lib.blackbox_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.blackbox_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absaroka_media_mount(self):
        """Operation to mount the absaroka repo partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_mount.argtypes = []
        lib.absaroka_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_mount.argtypes = []
        lib.gaius_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_mount.argtypes = []
        lib.vid_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def mountflat(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.flat_media_mount.argtypes = []
        lib.flat_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.flat_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.media_unmount.argtypes = []
        lib.media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def bb_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.bb_media_unmount.argtypes = []
        lib.bb_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.bb_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_unmount.argtypes = []
        lib.vid_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_unmount.argtypes = []
        lib.gaius_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absr_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_unmount.argtypes = []
        lib.absaroka_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def _send_receive_raw(self, indata):
        """Send and receive raw function for blob operations

        :param indata: The data to be sent to blob operation.
        :type indata: str.

        """
        excp = None
        for _ in range(0, 3):  # channel loop for iLO
            try:
                resp = self.channel.send_receive_raw(indata, 10)
                return resp
            except Exception as exp:
                self.channel.close()
                lib = self.gethprestchifhandle()
                self.channel = HpIlo(dll=lib)
                excp = exp
        if excp:
            raise excp

    def cert_login(self, cert_file, priv_key, key_pass):
        lib = self.gethprestchifhandle()
        lib.login_cert.argtypes = [c_void_p, c_char_p, c_char_p, c_char_p]
        lib.login_cert.restype = c_char_p

        token = lib.login_cert(self.channel.fhandle, cert_file, priv_key, key_pass)
        return token

    @staticmethod
    def gethprestchifhandle():
        """Multi platform handle for chif hprest library"""
        excp = None
        libhandle = None
        if os.name != "nt":
            libpath = "/opt/ilorest/lib64/libilorestchif.so"
            if os.path.isfile(libpath):
                libhandle = cdll.LoadLibrary(libpath)
        if not libhandle:
            libnames = (
                ["ilorest_chif.dll"]
                if os.name == "nt"
                else [
                    "ilorest_chif_dev.so",
                    "ilorest_chif.so",
                ]
            )
            for libname in libnames:
                try:
                    libpath = BlobStore2.checkincurrdirectory(libname)
                    # LOGGER.debug("Loading Library %s for libhpsrv", libpath)
                    libhandle = cdll.LoadLibrary(libpath)
                    if libhandle:
                        # LOGGER.debug("Got Libhandle %s for libhpsrv", libhandle)
                        break
                except Exception as exp:
                    excp = exp

        if libhandle:
            BlobStore2.setglobalhprestchifrandnumber(libhandle)
            return libhandle
        else:
            import site
            site_packages = site.getsitepackages()
            for package in site_packages:
                try:
                    if os.name != "nt":
                        libpath = os.path.join(package, "ilorest", "chiflibrary", "ilorest_chif.so")
                    else:
                        libpath = os.path.join(package, "ilorest", "chiflibrary", "ilorest_chif.dll")
                    libhandle = cdll.LoadLibrary(libpath)
                except Exception as exp:
                    excp = exp
                if libhandle:
                    BlobStore2.setglobalhprestchifrandnumber(libhandle)
                    return libhandle
        raise ChifDllMissingError(excp)

    @staticmethod
    def setglobalhprestchifrandnumber(libbhndl):
        """Set the random number for the chif handle
        :param libbhndl: The library handle provided by loading the chif library.
        :type libbhndl: library handle.
        """
        rndval = random.randint(1, 65535)
        libbhndl.updaterandval.argtypes = [c_ushort]
        libbhndl.updaterandval(rndval)

    @staticmethod
    def initializecreds(username=None, password=None, log_dir=None):
        """Get chif ready to use high security
        :param username: The username to login.
        :type username: str.
        :param password: The password to login.
        :type password: str.
        """

        dll = BlobStore2.gethprestchifhandle()
        if LOGGER.isEnabledFor(logging.DEBUG):
            logdir_c = create_string_buffer(log_dir.encode('utf-8'))
            dll.enabledebugoutput(logdir_c)
        dll.ChifInitialize(None)
        if username:
            if not password:
                return False
            if dll.ChifIsSecurityRequired() > 0:
                dll.initiate_credentials.argtypes = [c_char_p, c_char_p]
                dll.initiate_credentials.restype = POINTER(c_ubyte)

                usernew = create_string_buffer(username.encode("utf-8"))
                passnew = create_string_buffer(password.encode("utf-8"))

                dll.initiate_credentials(usernew, passnew)
                credreturn = dll.ChifVerifyCredentials()
                if not credreturn == BlobReturnCodes.SUCCESS:
                    if credreturn == hpiloreturncodes.CHIFERR_AccessDenied:
                        raise Blob2SecurityError()
                    else:
                        raise HpIloInitialError(
                            "Error %s occurred while trying to open a channel to iLO" % credreturn
                        )
            else:
                dll.ChifDisableSecurity()
        else:
            # if high security return False
            # LOGGER.debug("Calling ChifIsSecurityRequired...")
            if dll.ChifIsSecurityRequired() > 0:
                return False
            else:
                # LOGGER.debug("Calling ChifDisableSecurity...")
                dll.ChifDisableSecurity()

        BlobStore2.unloadchifhandle(dll)

        return True

    @staticmethod
    def checkincurrdirectory(libname):
        """Check if the library is present in current directory.
        :param libname: The name of the library to search for.
        :type libname: str."""
        libpath = libname

        if os.path.isfile(os.path.join(os.path.split(sys.executable)[0], libpath)):
            libpath = os.path.join(os.path.split(sys.executable)[0], libpath)
        elif os.path.isfile(os.path.join(os.getcwd(), libpath)):
            libpath = os.path.join(os.getcwd(), libpath)
        elif os.environ.get("LD_LIBRARY_PATH"):
            paths = os.getenv("LD_LIBRARY_PATH", libpath).split(";")
            libpath = [os.path.join(pat, libname) for pat in paths if os.path.isfile(os.path.join(pat, libname))]
            libpath = libpath[0] if libpath else libname

        return libpath

    @staticmethod
    def unloadchifhandle(lib):
        """Release a handle on the chif iLOrest library

        :param lib: The library handle provided by loading the chif library.
        :type lib: library handle.

        """
        try:
            libhandle = lib._handle
            if os.name == "nt":
                windll.kernel32.FreeLibrary(None, handle=libhandle)
            else:
                dlclose(libhandle)
        except Exception:
            pass



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of manually managing sessions with Redfish
"""

import sys
import json
from six.moves import urllib
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

def sessions(_redfishobj, login_account, login_password):

    new_session = {"UserName": login_account, "Password": login_password}
    response = _redfishobj.post('/redfish/v1/Sessions', new_session)

    if response.status == 201:
        print("Success!\n")
        session_uri = response.getheader("location")
        session_uri = urllib.parse.urlparse(session_uri)
        sys.stdout.write("\tSession " + session_uri.path + " created.\n")
        x_auth_token = response.getheader("x-auth-token")
        sys.stdout.write("\tThis is the session X-Auth Token key " + x_auth_token + ".\n")
        print(json.dumps(response.dict, indent=4, sort_keys=True))

        # Delete the created session
        sys.stdout.write("\tTerminating this session.\n")
        sessresp = _redfishobj.delete(session_uri.path)
        print(json.dumps(sessresp.dict, indent=4, sort_keys=True))
    else:
        sys.stderr.write("ERROR: failed to create a session.\n")
        try:
            print(json.dumps(sessresp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                            sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended"
                             " Message Info...")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    sessions(REDFISHOBJ, LOGIN_ACCOUNT, LOGIN_PASSWORD)



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from urllib3.util import parse_url
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def sessions(restobj, login_account, login_password):
    new_session = {"UserName": login_account, "Password": login_password}
    response = restobj.post("/rest/v1/Sessions", new_session)
    sys.stdout.write("%s" % response)

    if response.status == 201:
        session_uri = response.getheader("location")
        session_uri = parse_url(session_uri)
        sys.stdout.write("\tSession " + session_uri.path + " created\n")

        x_auth_token = response.getheader("x-auth-token")
        sys.stdout.write("\tSession key " + x_auth_token + " created\n")

        # Delete the created session
        sessresp = restobj.delete(session_uri.path)
        sys.stdout.write("%s" % response)
    else:
        sys.stderr.write("ERROR: failed to create a session.\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    sys.stdout.write("\nEXAMPLE 14: Create/Use/Delete a user session\n")
    sessions(REST_OBJ, "admin", "admin123")



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def computer_details(restobj):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "ComputerSystem." in instance.Type:
                system_path = instance.href
                break

    response = restobj.get(system_path)

    sys.stdout.write("\tManufacturer:  " + \
                            str(response.dict["Manufacturer"]) + "\n")
    sys.stdout.write("\tModel:  " + str(response.dict["Model"]) + "\n")
    sys.stdout.write("\tSerial Number:  " + \
                            str(response.dict["SerialNumber"]) + "\n")
    if "VirtualSerialNumber" in response.dict:
        sys.stdout.write("\tVirtual Serial Number:  " +
               str(response.dict["VirtualSerialNumber"]) + "\n")
    else:
        sys.stderr.write("\tVirtual Serial Number information not " \
                                    "available on system resource\n")
    sys.stdout.write("\tUUID:  " + str(response.dict["UUID"]) + "\n")

    if "VirtualUUID" in response.dict["Oem"]["Hp"]:
        sys.stdout.write("\tVirtualUUID:  " + \
                 str(response.dict["Oem"]["Hp"]["VirtualUUID"]) + "\n")
    else:
        sys.stderr.write("\tVirtualUUID not available system " \
                                                        "resource\n")
    if "AssetTag" in response.dict:
        sys.stdout.write("\tAsset Tag:  " + response.dict["AssetTag"] \
                                                                + "\n")
    else:
        sys.stderr.write("\tNo Asset Tag information on system " \
                                                        "resource\n")
    sys.stdout.write("\tBIOS Version: " + \
             response.dict["Bios"]["Current"]["VersionString"] + "\n")

    sys.stdout.write("\tMemory:  " +
           str(response.dict["Memory"]["TotalSystemMemoryGB"]) +" GB\n")

    sys.stdout.write("\tProcessors:  " + \
             str(response.dict["Processors"]["Count"]) + " x " + \
             str(response.dict["Processors"]["ProcessorFamily"])+ "\n")

    if "Status" not in response.dict or "Health" not in \
                                                response.dict["Status"]:
        sys.stdout.write("\tStatus/Health information not available in "
                                                    "system resource\n")
    else:
        sys.stdout.write("\tHealth:  " + \
                         str(response.dict["Status"]["Health"]) + "\n")

    if "HostCorrelation" in response.dict:
        if "HostFQDN" in response.dict["HostCorrelation"]:
            sys.stdout.write("\tHost FQDN:  " + \
                 response.dict["HostCorrelation"]["HostFQDN"] + "\n")

        if "HostMACAddress" in response.dict["HostCorrelation"]:
            for mac in response.dict["HostCorrelation"]["HostMACAddress"]:
                sys.stdout.write("\tHost MAC Address:  " + str(mac) + "\n")

        if "HostName" in response.dict["HostCorrelation"]:
            sys.stdout.write("\tHost Name:  " + \
                 response.dict["HostCorrelation"]["HostName"] + "\n")

        if "IPAddress" in response.dict["HostCorrelation"]:
            for ip_address in response.dict["HostCorrelation"]["IPAddress"]:
                if ip_address:
                    sys.stdout.write("\tHost IP Address:  " + str(ip_address) + "\n")

    sys.stdout.write("%s" % response)

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 16: Dump host computer details\n")
    computer_details(REST_OBJ)
    REST_OBJ.logout()



# Copyright 2020 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# -*- coding: utf-8 -*-
"""
An example of set syslog server on HPE iLO Server.
"""
import sys
import json
import argparse
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
global DISABLE_RESOURCE_DIR
from ilorest_util import get_resource_directory
from ilorest_util import get_gen

def set_syslog(_redfishobj, syslog_server):

    model_uri = "/redfish/v1/Systems/1/"
    model = _redfishobj.get(model_uri).obj['Model']
    print (model)
    if "Gen9" in model or 'Gen8' in model:
        hp = "Hp"
    else:
        hp = "Hpe"

    syslog_uri = "/redfish/v1/Managers/1/NetworkService/"

    body = {"Oem": {hp: {"RemoteSyslogServer": syslog_server, "RemoteSyslogEnabled": True}}}
    resp = _redfishobj.patch(syslog_uri, body)
    ilo_response(_redfishobj, resp)

def ilo_response(_redfishobj, resp):

    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                         sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended "\
                             "Message Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("Success")

if __name__ == "__main__":
    # Initialize parser
    parser = argparse.ArgumentParser(description = "Script to upload and flash NVMe FW")

    parser.add_argument(
        '-i',
        '--ilo',
        dest='ilo_ip',
        action="store",
        help="iLO IP of the server",
        default=None)
    parser.add_argument(
        '-u',
        '--user',
        dest='ilo_user',
        action="store",
        help="iLO username to login",
        default=None)
    parser.add_argument(
        '-p',
        '--password',
        dest='ilo_pass',
        action="store",
        help="iLO password to log in.",
        default=None)
    parser.add_argument(
        '-s',
        '--syslog_server',
        dest='syslog_server',
        action="store",
        required=True,
        help="iLO password to log in.",
        default=None)

    options = parser.parse_args()

    system_url = "https://" + options.ilo_ip
    print (system_url)

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = True

    try:
        # Create a Redfish client object
        redfish_obj = RedfishClient(base_url=system_url, username=options.ilo_user, password=options.ilo_pass)
        # Login with the Redfish client
        redfish_obj.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    (ilogen,_) = get_gen(redfish_obj)
    print ("Generation is ", ilogen)
    if int(ilogen) == 5:
        set_syslog(redfish_obj, options.syslog_server)
    else:
        set_syslog(redfish_obj, options.syslog_server)
    redfish_obj.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""RMC helper file. Includes RMC errors and caching functionality for monolith."""

# ---------Imports---------

import errno
import hashlib
import json
import logging
import os

from redfish.rest.containers import RestRequest, StaticRestResponse
from redfish.rest.v1 import RestClient

from .ris import RisMonolith
from .sharedtypes import JSONEncoder

# ---------End of imports---------

# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class RdmcError(Exception):
    """Base class for all RDMC Exceptions"""

    errcode = 1

    def __init__(self, message):
        Exception.__init__(self, message)


class InvalidCommandLineError(RdmcError):
    """Raised when user enter incorrect command line arguments"""

    pass

class TfaEnablePreRequisiteError(RdmcError):
    """Raised when pre-requisites not met while enabling TFA"""

    pass

class FailureDuringCommitError(RdmcError):
    """Raised when there is an error while committing."""

    pass


class UserNotAdminError(RdmcError):
    """Raised when user doesn't have admin priviledges, but they are required."""

    pass


class IncompatibleiLOVersionError(RdmcError):
    """Raised when iLO version is incompatible."""

    pass


class UndefinedClientError(Exception):
    """Raised when there are no clients active (usually when user hasn't logged in)."""

    pass


class InstanceNotFoundError(Exception):
    """Raised when attempting to select an instance that does not exist."""

    pass


class CurrentlyLoggedInError(Exception):
    """Raised when attempting to select an instance that does not exist"""

    pass


class IloLicenseError(Exception):
    """Raised when the proper iLO license is not available for a command"""

    pass


class ScepenabledError(Exception):
    """Raised when the generation csr or deletion of https cert is issues when scep is enabled"""

    pass


class NothingSelectedError(Exception):
    """Raised when attempting to access an object without first selecting it."""

    pass


class NothingSelectedFilterError(Exception):
    """Raised when the filter applied doesn't match any selection (general)."""

    pass


class NothingSelectedSetError(Exception):
    """Raised when attempting to access an object without first selecting it (In set)."""

    pass


class InvalidSelectionError(Exception):
    """Raised when selection argument fails to match anything."""

    pass


class IdTokenError(Exception):
    """Raised when user is not authorized to complete the operation."""

    pass


class ValueChangedError(Exception):
    """Raised if user tries to set/commit a value when monolith has older data."""

    pass


class LoadSkipSettingError(Exception):
    """Raised when one or more settings are absent in given server."""

    pass


class InvalidPathError(Exception):
    """Raised when requested path is not found."""

    pass


class UnableToObtainIloVersionError(Exception):
    """Raised when iloversion is missing from default path."""

    pass


class ValidationError(Exception):
    """Raised when there is a problem with user input."""

    def __init__(self, errlist):
        super(ValidationError, self).__init__(errlist)
        self._errlist = errlist

    def get_errors(self):
        """Returns error list."""
        return self._errlist


class IloResponseError(Exception):
    """Raised when iLO returns with a non 2XX response."""

    pass


class EmptyRaiseForEAFP(Exception):
    """Raised when you need to check for issues and take different action."""

    pass


class IncorrectPropValue(Exception):
    """Raised when you pass an incorrect value to for the associated property."""

    pass


class RmcCacheManager(object):
    """Manages caching/uncaching of data for RmcApp.

    :param rmc: RmcApp to be managed
    :type rmc: :class:`redfish.ris.rmc.RmcApp`
    """

    def __init__(self, rmc):
        """Initialize RmcCacheManager

        :param rmc: RmcApp to be managed
        :type rmc: RmcApp object

        """
        self._rmc = rmc

        self.encodefunct = lambda data: data
        self.decodefunct = lambda data: data


class RmcFileCacheManager(RmcCacheManager):
    """RMC file cache manager.

    :param rmc: RmcApp to be managed
    :type rmc: :class:`redfish.ris.rmc.RmcApp`
    """

    def __init__(self, rmc):
        super(RmcFileCacheManager, self).__init__(rmc)

    def logout_del_function(self, url=None):
        """Searches for a specific url in cache or returns all urls and returns them for RmcApp
        to run logout on, clearing the session.

        :param url: The URL to pass back for logout.
        :type url: str
        """
        if self._rmc.cache:
            cachedir = self._rmc.cachedir
            indexfn = os.path.join(cachedir, "index")  # %s\\index' % cachedir
        else:
            indexfn = ""
        sessionlocs = []

        if os.path.isfile(indexfn):
            try:
                indexfh = open(indexfn, "r")
                index_cache = json.load(indexfh)
                indexfh.close()

                for index in index_cache:
                    if url:
                        if url in index["url"]:
                            os.remove(os.path.join(cachedir, index["href"]))
                            break
                    else:
                        if os.path.isfile(os.path.join(cachedir, index["href"])):
                            monolith = open(os.path.join(cachedir, index["href"]), "r")
                            data = json.load(monolith)
                            monolith.close()
                            for item in data:
                                if "login" in item and "session_location" in data["login"]:
                                    if "blobstore" in data["login"]["url"]:
                                        loc = data["login"]["session_location"].split("//")[-1]
                                        sesurl = None
                                    else:
                                        loc = None
                                        if data["login"]["session_location"] is not None:
                                            loc = data["login"]["session_location"].split(data["login"]["url"])[-1]
                                        sesurl = data["login"]["url"]
                                    sessionlocs.append(
                                        (
                                            loc,
                                            sesurl,
                                            self._rmc._cm.decodefunct(data["login"]["session_key"]),
                                        )
                                    )

                        os.remove(os.path.join(cachedir, index["href"]))
            except BaseException as excp:
                LOGGER.warning("Unable to read cache data %s", excp)

        return sessionlocs

    def uncache_rmc(self, creds=None, enc=False):
        """Uncaches monolith data from cache location specified by RmcApp.

        :param creds: Dictionary of username and password.
                      Only required for restoring high security local calls.
        :type creds: dict
        :param enc: Flag if credentials passed are encoded.
        :type enc: bool
        """
        cachedir = self._rmc.cachedir
        indexfn = "%s/index" % cachedir

        if os.path.isfile(indexfn):
            try:
                indexfh = open(indexfn, "r")
                index_cache = json.load(indexfh)
                indexfh.close()

                for index in index_cache:
                    clientfn = index["href"]
                    self._uncache_client(clientfn, creds=creds, enc=enc)
            except BaseException as excp:
                LOGGER.warning("Unable to read cache data %s", excp)

    def _uncache_client(self, cachefn, creds=None, enc=False):
        """Monolith uncache function for parsing and passing all client data and associated
           credential attributes.

        :param cachefn: The cache file name.
        :type cachefn: str.

        """
        cachedir = self._rmc.cachedir
        clientsfn = "%s/%s" % (cachedir, cachefn)

        if os.path.isfile(clientsfn):
            try:
                clientsfh = open(clientsfn, "r")
                client = json.load(clientsfh)
                clientsfh.close()

                if "login" not in client:
                    return

                login_data = client["login"]
                if "url" not in login_data:
                    return

                self._rmc.typepath.getgen(
                    login_data.get("ilo"),
                    url=login_data.get("url"),
                    isredfish=login_data.get("redfish", None),
                    ca_cert_data=login_data.get("ca_cert_data", {}),
                )

                if creds and login_data.get("url", "").startswith("blobstore://"):
                    if enc:
                        creds["username"] = self._rmc._cm.decodefunct(creds["username"])
                        creds["password"] = self._rmc._cm.decodefunct(creds["password"])
                    login_data["username"] = creds["username"]
                    login_data["password"] = creds["password"]
                    if isinstance(login_data["username"], bytes):
                        login_data["username"] = login_data["username"].decode("utf-8")
                    if isinstance(login_data["password"], bytes):
                        login_data["password"] = login_data["password"].decode("utf-8")

                redfishinst = RestClient(
                    username=login_data.get("username", "Administrator"),
                    password=login_data.get("password", None),
                    base_url=login_data.get("url", None),
                    biospassword=login_data.get("bios_password", None),
                    is_redfish=login_data.get("redfish", None),
                    default_prefix=self._rmc.typepath.defs.startpath,
                    proxy=login_data.get("proxy", None),
                    ca_cert_data=login_data.get("ca_cert_data", {}),
                )
                if login_data.get("authorization_key"):
                    redfishinst.basic_auth = login_data.get("authorization_key")
                elif login_data.get("session_key"):
                    redfishinst.session_key = self._rmc._cm.decodefunct(login_data.get("session_key"))
                    if isinstance(redfishinst.session_key, bytes):
                        redfishinst.session_key = redfishinst.session_key.decode("utf-8")
                    redfishinst.session_location = login_data.get("session_location")

                if "selector" in client:
                    self._rmc.selector = client["selector"]
                if login_data.get("iloversion"):
                    redfishinst.iloversion = login_data.get("iloversion")
                else:
                    redfishinst.iloversion = None
                self._rmc.typepath.iloversion = redfishinst.iloversion

                getdata = client["get"]
                for key in list(getdata.keys()):
                    if key == redfishinst.default_prefix:
                        restreq = RestRequest(method="GET", path=key)
                        getdata[key]["restreq"] = restreq
                        redfishinst.root = StaticRestResponse(**getdata[key])
                        break

                self._rmc.monolith = RisMonolith(redfishinst, self._rmc.typepath)
                self._rmc.monolith.load_from_dict(client["monolith"])
                self._rmc.redfishinst = redfishinst
                # make sure root is there
                _ = redfishinst.root
                self._rmc.typepath.defineregschemapath(redfishinst.root.dict)
            except BaseException as excp:
                LOGGER.warning("Unable to read cache data %s", excp)

    def cache_rmc(self):
        """Saves monolith data to the file path specified in RmcApp."""
        if not self._rmc.cache:
            return

        cachedir = self._rmc.cachedir
        if not os.path.isdir(cachedir):
            try:
                os.makedirs(cachedir)
            except OSError as ex:
                if ex.errno == errno.EEXIST:
                    pass
                else:
                    raise

        index_map = dict()
        index_cache = list()

        if self._rmc.redfishinst:
            shaobj = hashlib.new("SHA256")
            shaobj.update(self._rmc.redfishinst.base_url.encode("utf-8"))
            md5str = shaobj.hexdigest()

            index_map[self._rmc.redfishinst.base_url] = md5str
            index_data = dict(
                url=self._rmc.redfishinst.base_url,
                href="%s" % md5str,
            )
            index_cache.append(index_data)

            indexfh = open("%s/index" % cachedir, "w")
            json.dump(index_cache, indexfh, indent=2, cls=JSONEncoder)
            indexfh.close()

        if self._rmc.redfishinst:
            login_data = dict(
                username=None,
                password=None,
                url=self._rmc.redfishinst.base_url,
                session_key=self._rmc._cm.encodefunct(self._rmc.redfishinst.session_key),
                session_location=self._rmc.redfishinst.session_location,
                authorization_key=self._rmc.redfishinst.basic_auth,
                bios_password=self._rmc.redfishinst.bios_password,
                redfish=self._rmc.monolith.is_redfish,
                ilo=self._rmc.typepath.ilogen,
                iloversion=self._rmc.typepath.iloversion,
                proxy=self._rmc.redfishinst.proxy,
                ca_cert_data=self._rmc.redfishinst.connection._connection_properties
                if self._rmc.redfishinst.connection._connection_properties
                else dict(),
            )

            clients_data = dict(
                selector=self._rmc.selector,
                login=login_data,
                monolith=self._rmc.monolith,
                get=self._rmc.monolith.paths,
            )

            clientsfh = open("%s/%s" % (cachedir, index_map[self._rmc.redfishinst.base_url]), "w")

            if isinstance(clients_data, bytes):
                clients_data = clients_data.decode("utf-8")
            json.dump(clients_data, clientsfh, indent=2, cls=JSONEncoder)
            clientsfh.close()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Compatibility functionality in between iLO versions and generic Redfish/LegacyRest servers.
Used to provide convenient string variables that are usable on any iLO irrespective of version or
API type."""
# ---------Imports---------
import logging

from redfish import LegacyRestClient, RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
from redfish.ris.rmc_helper import (
    NothingSelectedError,
    UnableToObtainIloVersionError,
    UserNotAdminError,
)

# ---------End of imports---------

LOGGER = logging.getLogger(__name__)


# TODO: This will be a part of the compatability class
class Typesandpathdefines(object):
    """The global types and path definitions class. Holds information on a system and automatically
    creates the correct type strings along with some generic paths. Paths are meant to be used with
    iLO systems. Paths may be different on generic Redfish systems. Self variables are created when
    the `getgen` function is called.

    Useful self variables that are created include:

    * **url**: The url of the system that the defines were created for.
    * **defs**: The string defines for the system that was passed to `getgen`. Includes key
      property keys, paths, types, and flags to check what the system type is.
    * **ilogen**: The iLO generation of the system that the defines were created for. For non-iLO
      Redfish systems this is set to **5**.
    * **iloversion**: The iLO version of the system that the defines were created for.
    * **flagiften**: Flag is set to true if the system is Gen 10 or a non-iLO Redfish system.
    """

    def __init__(self):
        self.url = None
        self.defs = None
        self.ilogen = None
        self.iloversion = None
        self.flagiften = False
        self.adminpriv = True

    def getgen(
        self,
        gen=None,
        url=None,
        username=None,
        password=None,
        sessionid=None,
        logger=None,
        proxy=None,
        ca_cert_data={},
        isredfish=True,
        login_otp=None,
        log_dir=None,
    ):
        """Function designed to verify the servers platform. Will generate the `Typeandpathdefines`
        variables based on the system type that is detected.

        :param url: The URL to perform the request on.
        :type url: str
        :param username: The username to login with.
        :type username: str
        :param password: The password to login with.
        :type password: str
        :param proxy: The proxy to connect to the system with.
        :type proxy: str
        :param ca_certs: Dictionary including the TLS certificate information of user based
          authentication
        :type ca_certs: dict
        :param isredfish: The flag to force redfish conformance on iLO 4 systems. You will still
          need to call `updatedefinesflag` to update the defines to Redfish.
        :type isredfish: bool
        :param logger: The logger handler to log data too uses the default if none is provided.
        :type logger: str
        """

        if self.adminpriv is False and url.startswith("blob"):
            raise UserNotAdminError("")

        self.url = url
        self.is_redfish = isredfish
        self.gencompany = self.rootresp = False
        self.ilogen = 5  # If no iLO or Anonymous data , default to iLO 5 types
        logger = logger if not logger else LOGGER
        client = None
        self.noschemas = False
        self.schemapath = self.regpath = ""

        if not gen:
            try_count = 0
            try:
                client = RedfishClient(
                    base_url=self.url,
                    username=username,
                    password=password,
                    sessionid=sessionid,
                    proxy=proxy,
                    ca_cert_data=ca_cert_data,
                    login_otp=login_otp,
                    log_dir=log_dir,
                )
                client._get_root()
            except ServerDownOrUnreachableError as excp:
                if self.is_redfish:
                    raise excp
                try_count += 1
            if not self.is_redfish:
                try:
                    restclient = LegacyRestClient(
                        base_url=self.url,
                        username=username,
                        password=password,
                        sessionid=sessionid,
                        proxy=proxy,
                        ca_cert_data=ca_cert_data,
                        login_otp=login_otp
                    )
                    restclient._get_root()
                    # Check that the response is actually legacy rest and not a redirect
                    _ = restclient.root.obj.Type
                    self.is_redfish = False
                    client = restclient
                except Exception as excp:
                    try_count += 1
                    if not client:
                        logger.info("Gen get rest error:" + str(excp) + "\n")
                        raise excp
                    else:
                        self.is_redfish = True

            if try_count > 1:
                raise ServerDownOrUnreachableError(
                    "Server not reachable or does not support " "HPRest or Redfish \n"
                )

            rootresp = client.root.obj
            self.rootresp = rootresp
            client.logout()

            self.gencompany = next(iter(self.rootresp.get("Oem", {}).keys()), None) in (
                "Hpe",
                "Hp",
            )
            comp = "Hp" if self.gencompany else None
            comp = "Hpe" if rootresp.get("Oem", {}).get("Hpe", None) else comp
            if comp and next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get("ManagerType", None):
                self.ilogen = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get("ManagerType")
                self.ilover = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get(
                    "ManagerFirmwareVersion"
                )
                if self.ilogen.split(" ")[-1] == "CM":
                    # Assume iLO 4 types in Moonshot
                    self.ilogen = 4
                    self.iloversion = None
                else:
                    self.iloversion = float(self.ilogen.split(" ")[-1] + "." + "".join(self.ilover.split(".")))
        else:
            self.ilogen = int(gen)

        try:
            if not isinstance(self.ilogen, int):
                self.ilogen = int(self.ilogen.split(" ")[-1])
            self.flagiften = True if int(self.ilogen) >= 5 else False
        except:
            raise UnableToObtainIloVersionError("Unable to find the iLO generation.")

        self.noschemas = (
            True
            if self.rootresp and "JsonSchemas" in self.rootresp and not self.rootresp.get("JsonSchemas", None)
            else False
        )
        if self.noschemas:
            self.ilogen = self.ilover = self.iloversion = None
        if self.rootresp and not self.noschemas:
            self.defineregschemapath(self.rootresp)

        if self.flagiften:
            self.defs = Definevalstenplus()
        else:
            self.defs = DefinevalsNine()

    def defineregschemapath(self, rootobj):
        """Defines the schema and registry paths using data in root path.

        :param rootobj: The root path data.
        :type rootobj: dict.
        """
        self.gencompany = next(iter(rootobj.get("Oem", {}).keys()), None) in ("Hpe", "Hp")
        self.schemapath = (
            rootobj["JsonSchemas"]["@odata.id"]
            if rootobj.get("JsonSchemas", None)
            else rootobj["links"]["Schemas"]["href"]
        )
        self.schemapath = (
            self.schemapath.rstrip("/") + "/?$expand=."
            if self.is_redfish and self.flagiften and self.gencompany
            else self.schemapath
        )
        self.regpath = (
            rootobj["Registries"]["@odata.id"]
            if rootobj.get("Registries", None)
            else rootobj["links"]["Registries"]["href"]
        )
        self.regpath = (
            self.regpath.rstrip("/") + "/?$expand=."
            if self.is_redfish and self.flagiften and self.gencompany
            else self.regpath
        )

    # TODO: Move these to a compatability class
    def updatedefinesflag(self, redfishflag=None):
        """Updates the redfish and rest flag depending on system and redfishflag input. On an iLO 5
        system or another Redfish system, this will do nothing. On an iLO 4 system with both Redfish
        and LegacyRest this will update the defines to redfish if the *redfishflag* is set to True
        and stay with the LegacyRest defines otherwise.

        :param redfishflag: User input for redfish
        :type redfishflag: bool
        :returns: True if the system should use Redfish, False for legacy Rest.
        :rtype: bool
        """
        if self.defs:
            is_redfish = redfishflag or self.defs.isgen10
            self.defs.flagforrest = not is_redfish
            if is_redfish:
                self.defs.redfishchange()

            return is_redfish
        else:
            return redfishflag

    def modifyselectorforgen(self, sel):
        """Changes the select to match the Generation's HP string based to the correct type for
        specific iLO versions.

        :param sel: query to be changed to match Generation's HP string
        :type sel: str
        :returns: A modified selector matching the Generation's HP string.
        :rtype: string
        """
        if not sel:
            raise NothingSelectedError()
        sel = sel.lower()
        returnval = sel

        if sel.startswith(("hpeeskm", "#hpeeskm", "hpeskm", "#hpeskm")):
            returnval = self.defs.hpeskmtype
        elif "bios." in sel[:9].lower():
            returnval = self.defs.biostype
        elif sel.startswith(("hpe", "#hpe")) and self.defs and self.defs.isgen9:
            returnval = sel[:4].replace("hpe", "hp") + sel[4:]
        elif not sel.startswith(("hpe", "#hpe")) and self.defs and self.defs.isgen10:
            returnval = sel[:3].replace("hp", "hpe") + sel[3:]

        return returnval


class Definevals(object):
    """Base class for setting platform dependent variables."""

    def __init__(self):
        pass


class Definevalstenplus(Definevals):
    """Platform dependent variables for iLO 5+ (Gen 10)."""

    # pylint: disable=too-many-instance-attributes
    # As a defines class this will need all the attributes
    def __init__(self):
        self.oemhp = "Hpe"

        self.oempath = "/Oem/Hpe"
        self.startpath = "/redfish/v1/"
        self.systempath = "/redfish/v1/Systems/1/"
        self.managerpath = "/redfish/v1/Managers/1/"
        self.biospath = "/redfish/v1/systems/1/bios/"
        self.addlicensepath = "/redfish/v1/Managers/1/LicenseService/"
        self.accountspath = "/redfish/v1/AccountService/Accounts/"
        self.federationpath = "/redfish/v1/Managers/1/FederationGroups/"
        self.resourcedirpath = "/redfish/v1/ResourceDirectory/"

        self.biostype = "Bios."
        self.hpeskmtype = "HpeESKM."
        self.hpcommontype = "HpeCommon"
        self.hpilossotype = "HpeiLOSSO."
        self.hpsecureboot = "SecureBoot."
        self.logservicetype = "#LogService."
        self.iscsisource = "iSCSISources"
        self.iscsiattemptinstance = "iSCSIAttemptInstance"
        self.iscsiattemptname = "iSCSIAttemptName"
        self.hphttpscerttype = "HpeHttpsCert."
        self.snmpservice = "HpeiLOSnmpService."
        self.attributenametype = "AttributeName"
        self.hpilodatetimetype = "HpeiLODateTime."
        self.attributeregtype = "#AttributeRegistry."
        self.hpilofirmwareupdatetype = "UpdateService."
        self.resourcedirectorytype = "HpeiLOResourceDirectory."
        self.hpilofederationgrouptype = "HpeiLOFederationGroup."
        self.managernetworkservicetype = "ManagerNetworkProtocol."
        self.schemafilecollectiontype = "#JsonSchemaFileCollection."
        self.regfilecollectiontype = "#MessageRegistryFileCollection."
        self.hpilolicensecollectiontype = "HpeiLOLicenseCollection."
        self.hpiloactivehealthsystemtype = "#HpeiLOActiveHealthSystem."
        self.securityservice = "HpeSecurityService."
        self.hpiscsisoftwareinitiatortype = "HpeiSCSISoftwareInitiator."
        self.hpilofederationgrouptypecoll = "HpeiLOFederationGroupCollection."
        self.bootoverridetargettype = "BootSourceOverrideTarget@Redfish.AllowableValues"
        self.messageregistrytype = "#MessageRegistry."

        self.typestring = "@odata.type"
        self.hrefstring = "@odata.id"
        self.collectionstring = "Members"
        self.biossettingsstring = "@Redfish.Settings"
        self.attname = "AttributeName"
        self.iscsistring = "iSCSISources"

        self.isgen9 = False
        self.isgen10 = True
        self.flagforrest = False
        super(Definevalstenplus, self).__init__()

    def redfishchange(self):
        """Empty function to update redfish variables (unneeded when the system is already redfish)."""
        pass


class DefinevalsNine(Definevals):
    """Platform dependent variables for iLO 4 LegacyRest (Gen 9)."""

    # pylint: disable=too-many-instance-attributes
    # As a defines class this will need all the attributes
    def __init__(self):
        self.oemhp = "Hp"

        self.oempath = "/Oem/Hp"
        self.startpath = "/rest/v1"
        self.systempath = "/rest/v1/Systems/1"
        self.managerpath = "/rest/v1/Managers/1"
        self.biospath = "/rest/v1/systems/1/bios"
        self.addlicensepath = "/rest/v1/Managers/1/LicenseService"
        self.accountspath = "/rest/v1/AccountService/Accounts"
        self.federationpath = "/rest/v1/Managers/1/FederationGroups"
        self.resourcedirpath = "/rest/v1/ResourceDirectory"

        self.biostype = "HpBios."
        self.hpeskmtype = "HpESKM."
        self.hpcommontype = "HpCommon"
        self.hpilossotype = "HpiLOSSO."
        self.snmpservice = "SnmpService."
        self.attributenametype = "Name"
        self.logservicetype = "LogService."
        self.iscsisource = "iSCSIBootSources"
        self.iscsiattemptinstance = "iSCSIBootAttemptInstance"
        self.iscsiattemptname = "iSCSIBootAttemptName"
        self.hpsecureboot = "HpSecureBoot."
        self.hphttpscerttype = "HpHttpsCert."
        self.hpilodatetimetype = "HpiLODateTime."
        self.hpilofirmwareupdatetype = "HpiLOFirmwareUpdate."
        self.resourcedirectorytype = "HpiLOResourceDirectory."
        self.hpilofederationgrouptype = "HpiLOFederationGroup."
        self.attributeregtype = "HpBiosAttributeRegistrySchema."
        self.schemafilecollectiontype = "#SchemaFileCollection."
        self.regfilecollectiontype = "#SchemaFileCollection."
        self.managernetworkservicetype = "ManagerNetworkService."
        self.hpiloactivehealthsystemtype = "HpiLOActiveHealthSystem."
        self.messageregistrytype = "MessageRegistry."
        self.hpilolicensecollectiontype = None
        self.hpilofederationgrouptypecoll = None
        self.bootoverridetargettype = "BootSourceOverrideSupported"
        self.hpiscsisoftwareinitiatortype = "HpiSCSISoftwareInitiator"

        self.typestring = "Type"
        self.hrefstring = "href"
        self.collectionstring = "Items"
        self.biossettingsstring = "SettingsResult"
        self.attname = "Name"
        self.iscsistring = "iSCSIBootSources"

        self.isgen9 = True
        self.isgen10 = False
        self.flagforrest = True
        super(DefinevalsNine, self).__init__()

    def redfishchange(self):
        """Function to update redfish variables from LegacyRest to iLO 4 Redfish (Gen 9)."""
        self.startpath = "/redfish/v1/"
        self.systempath = "/redfish/v1/Systems/1/"
        self.managerpath = "/redfish/v1/Managers/1/"
        self.biospath = "/redfish/v1/systems/1/bios/"
        self.addlicensepath = "/redfish/v1/Managers/1/LicenseService/"
        self.resourcedirpath = "/redfish/v1/ResourceDirectory/"

        self.typestring = "@odata.type"
        self.hrefstring = "@odata.id"
        self.collectionstring = "Members"

        self.logservicetype = "#LogService."
        self.hpiloactivehealthsystemtype = "#HpiLOActiveHealthSystem."
        self.hpilolicensecollectiontype = "HpiLOLicenseCollection."
        self.hpilofederationgrouptypecoll = "HpiLOFederationGroupCollection."
        self.managernetworkservicetype = "ManagerNetworkProtocol."

        self.flagforrest = False



###
# Copyright 2019 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Module for working with global configuration options."""

#---------Imports---------

import os
import re
import logging
import six
from six.moves import configparser

#---------End of imports---------


#---------Debug logger---------

LOGGER = logging.getLogger(__name__)

#---------End of debug logger---------

class AutoConfigParser(object):
    """Auto configuration parser"""
    # properties starting with _ac__ are automatically
    # serialized to config file
    _config_pattern = re.compile(r'_ac__(?P<confkey>.*)')

    def __init__(self, filename=None):
        """Initialize AutoConfigParser

        :param filename: file name to be used for config loading.
        :type filename: str.

        """
        self._sectionname = 'globals'
        self._configfile = filename

    def _get_ac_keys(self):
        """Retrieve parse option keys"""
        result = []
        for key in six.iterkeys(self.__dict__):
            match = AutoConfigParser._config_pattern.search(key)
            if match:
                result.append(match.group('confkey'))
        return result

    def _get(self, key):
        """Retrieve parse option key

        :param key: key to retrieve.
        :type key: str.

        """
        ackey = '_ac__%s' % key.replace('-', '_')
        if ackey in self.__dict__:
            return self.__dict__[ackey]
        return None

    def _set(self, key, value):
        """Set parse option key

        :param key: key to be set.
        :type key: str.
        :param value: value to be given to key.
        :type value: str.

        """
        ackey = '_ac__%s' % key.replace('-', '_')
        if ackey in self.__dict__:
            self.__dict__[ackey] = value
        return None

    def load(self, filename=None):
        """Load configuration settings from the file filename, if filename
        is None then the value from get_configfile() is used

        :param filename: file name to be used for config loading.
        :type filename: str.

        """
        fname = self.get_configfile()
        if filename:
            fname = filename

        if not fname or not os.path.isfile(fname):
            return

        try:
            config = configparser.RawConfigParser()
            config.read(fname)
            for key in self._get_ac_keys():
                configval = None
                try:
                    configval = config.get(self._sectionname, key)
                except configparser.NoOptionError:
                    # also try with - instead of _
                    try:
                        configval = config.get(self._sectionname, key.replace('_', '-'))
                    except configparser.NoOptionError:
                        pass

                if configval:
                    ackey = '_ac__%s' % key
                    self.__dict__[ackey] = configval
        except configparser.NoOptionError:
            pass
        except configparser.NoSectionError:
            pass

    def save(self, filename=None):
        #TODO: Maybe unused
        """Save configuration settings from the file filename, if filename
        is None then the value from get_configfile() is used

        :param filename: file name to be used for config saving.
        :type filename: str.

        """
        fname = self.get_configfile()
        if filename:
            fname = filename

        if fname:
            return

        config = configparser.RawConfigParser()
        try:
            config.add_section(self._sectionname)
        except configparser.DuplicateSectionError:
            pass # ignored

        for key in self._get_ac_keys():
            ackey = '_ac__%s' % key
            config.set(self._sectionname, key, str(self.__dict__[ackey]))

        fileh = open(self._configfile, 'wb')
        config.write(fileh)
        fileh.close()

    def get_configfile(self):
        """ The current configuration file location"""
        return self._configfile



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def get_powermetrics_average(restobj):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "PowerMetrics." in instance.Type:
                power_path = instance.href
                break

    response = restobj.get(power_path)

    if "PowerMetrics" not in response.dict or \
        "AverageConsumedWatts" not in response.dict["PowerMetrics"] or \
                    "IntervalInMin" not in response.dict["PowerMetrics"]:
        sys.stdout.write("\tPowerMetrics resource does not contain "\
                   "'AverageConsumedWatts' or 'IntervalInMin' property\n")
    else:
        sys.stdout.write("\t" + " AverageConsumedWatts = " + \
             str(response.dict["PowerMetrics"]["AverageConsumedWatts"]) + \
             " watts over a " + str(response.dict["PowerMetrics"]\
                            ["IntervalInMin"]) + " minute moving average\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 30: Report PowerMetrics Average Watts\n")
    get_powermetrics_average(REST_OBJ)
    REST_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of obtaining system schema data
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

def get_schema(_redfishobj):
    schema_members_uris = []
    schema_members_expanded = {}

    schema_uri = _redfishobj.root.obj['JsonSchemas']['@odata.id']
    schema_response = _redfishobj.get(schema_uri)
    schema_members_uris = schema_response.obj['Members']
    for member in schema_members_uris:
        data = _redfishobj.get(member['@odata.id']).dict
        instance_id = data['Location'].index(next(iter(data['Location'])))
        schema_uri = data['Location'][instance_id]['Uri']
        schema = _redfishobj.get(schema_uri).dict
        #I am just replacing the URI link for the sub-schema with the actual schema. You will
        #see EVERYTHING
        _tmp = {schema_uri + ' ->': schema}
        data['Location'][instance_id]['Uri'] = _tmp
        schema_members_expanded[member['@odata.id']] = data

    print(json.dumps(schema_members_expanded, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    get_schema(REDFISHOBJ)
    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def get_base_registry(restobj):
    response = restobj.get("/rest/v1/Registries")
    messages = {}

    identifier = None

    for entry in response.dict["Items"]:
        if "Id" in entry:
            identifier = entry["Id"]
        else:
            identifier = entry["Schema"].split(".")[0]

        if identifier not in ["Base", "iLO"]:
            continue

        for location in entry["Location"]:
            reg_resp = restobj.get(location["Uri"]["extref"])

            if reg_resp.status == 200:
                sys.stdout.write("\tFound " + identifier + " at " + \
                                            location["Uri"]["extref"] + "\n")
                messages[identifier] = reg_resp.dict["Messages"]
            else:
                sys.stdout.write("\t" + identifier + " not found at "\
                                            + location["Uri"]["extref"] + "\n")

    return messages

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 2: Find and return registry " + "\n")
    get_base_registry(REST_OBJ)
    REST_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient

def get_resource_directory(restobj):
    response = restobj.get("/rest/v1/resourcedirectory")
    resources = []

    if response.status == 200:
        resources = response.obj.Instances
    else:
        sys.stderr.write("\tResource directory missing at /rest/v1/resourcedirectory" + "\n")
    return resources

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 1: Find the resource directory " + "\n")
    resources = get_resource_directory(REST_OBJ)

    for resource in resources:
        try:
            sys.stdout.write("\t" + str(resource["@odata.type"]) + \
                             "\n\t\t" + str(resource["@odata.id"]) + "\n")
        except KeyError:
            pass

    REST_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def modify_ilo_user_account(restobj, ilo_login_name_to_modify, \
                new_ilo_loginname, new_ilo_username, new_ilo_password, \
                irc=None, cfg=None, virtual_media=None, usercfg=None, vpr=None):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "Collection." in instance.Type:
                if "ManagerAccount." in instance.MemberType:
                    accounts_path = instance.href
                    break
    accounts = restobj.get(accounts_path)
    for account in accounts.obj.Items:
        for account in accounts.dict["Items"]:
            if account["UserName"] == ilo_login_name_to_modify:
                body = {}
                body_oemhp = {}
                body_oemhp_privs = {}

                # if new loginname or password specified
                if new_ilo_password:
                    body["Password"] = new_ilo_password
                if new_ilo_loginname:
                    body["UserName"] = new_ilo_loginname

                # if different username specified
                if new_ilo_username:
                    body_oemhp["LoginName"] = new_ilo_username

                # if different privileges were requested (None = no change)
                if irc != None:
                    body_oemhp_privs["RemoteConsolePriv"] = irc
                if virtual_media != None:
                    body_oemhp_privs["VirtualMediaPriv"] = virtual_media
                if cfg != None:
                    body_oemhp_privs["iLOConfigPriv"] = cfg
                if usercfg != None:
                    body_oemhp_privs["UserConfigPriv"] = usercfg
                if vpr != None:
                    body_oemhp_privs["VirtualPowerAndResetPriv"] = vpr

                # component assembly
                if len(body_oemhp_privs):
                    body_oemhp["Privileges"] = body_oemhp_privs
                if len(body_oemhp):
                    body["Oem"] = {"Hp": body_oemhp}

                newrsp = restobj.patch(account["links"]["self"]["href"], body)
                sys.stdout.write("%s" % newrsp)
                return

    sys.stderr.write("Account not found\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 11: Modify an iLO user account\n")
    modify_ilo_user_account(REST_OBJ, "name", "newname", "newusername", "newpassword")
    REST_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of setting an ISCSI boot instance for HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def set_bios_iscsi(_redfishobj, iscsi_properties):

    iscsi_uri = None
    iscsi_data = None

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
        bios_uri = systems_members_response.obj['Bios']['@odata.id']
        bios_response = _redfishobj.get(bios_uri)
        iscsi_uri = bios_response.obj.Oem.Hpe.Links['iScsi']['@odata.id']
        iscsi_data = _redfishobj.get(iscsi_uri)
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#HpeiSCSISoftwareInitiator.' in instance['@odata.type']:
                iscsi_uri = instance['@odata.id']
                iscsi_data = _redfishobj.get(iscsi_uri)

    if iscsi_data:
        for indx, inst in enumerate(iscsi_data.obj['iSCSISources']):
            if iscsi_properties['iSCSIBootInstance'] == indx:
                iscsi_data.dict['iSCSISources'][indx].update(iscsi_properties)
                resp = _redfishobj.patch(iscsi_uri, {'iSCSISources' : iscsi_data.dict['iSCSISources']})
                #If iLO responds with soemthing outside of 200 or 201 then lets check the
                #iLO extended info error message to see what went wrong
                if resp.status == 400:
                    try:
                        print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], \
                                                                        indent=4, sort_keys=True))
                    except Exception as excp:
                        print(json.dumps(resp.ori, indent=4, sort_keys=True))
                        sys.stderr.write("A response error occurred, unable to access iLO " \
                                         "Extended Message Info...")
                elif resp.status != 200:
                    sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
                else:
                    print("Success! A system reboot will be required to complete the change.\n")
                    print(json.dumps(resp.dict, indent=4, sort_keys=True))
                break

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #iscsi properties
    ISCSI_PROPERTIES = {"iSCSIBootInstance": 2, \
                      "iSCSITargetName": "Target", \
                      "iSCSIAttemptName": "Empty", \
                      "iSCSIConnectRetry": 5}
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    set_bios_iscsi(REDFISHOBJ, ISCSI_PROPERTIES)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
"""Monolith database implementation. Crawls Redfish and Legacy REST implementations
   and holds all data retrieved. The created database is called the **monolith** and referenced as
   such in the code documentation."""

# ---------Imports---------

import logging
import re
import sys
import threading
from re import error as regexerr

# Added for py3 compatibility
import six

try:
    from collections import OrderedDict, defaultdict
except ImportError:
    from collections.abc import OrderedDict, defaultdict

from queue import Queue

import jsonpath_rw
import jsonpointer
from jsonpointer import set_pointer
from six.moves.urllib.parse import urlparse, urlunparse
from six.moves.urllib.parse import ParseResult
from six.moves.urllib.parse import quote

from redfish.rest.containers import RestRequest, StaticRestResponse
from redfish.ris.ris_threaded import LoadWorker
from redfish.ris.sharedtypes import Dictable

# ---------End of imports---------

# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class BiosUnregisteredError(Exception):
    """Raised when BIOS has not been registered correctly in iLO"""

    pass


class SchemaValidationError(Exception):
    """Schema Validation Class Error"""

    pass


class SessionExpired(Exception):
    """Raised when session has expired"""

    pass


class RisMonolithMemberBase(Dictable):
    """RIS monolith member base class"""

    pass


class RisInstanceNotFoundError(Exception):
    """Raised when attempting to select an instance that does not exist"""

    pass


class RisMonolithMemberv100(RisMonolithMemberBase):
    """Class used by :class:`RisMonolith` for holding information on a response and adds extra data
    for monolith usage. A member can be marked as *modified* which means another operation may have
    rendered this member out of date. It should be reloaded before continuing to ensure data is
    up to date.

    :param restresp: `RestResponse` to create a member from.
    :type restresp: :class:`redfish.rest.containers.RestResponse`
    :param isredfish: Flag if the response is redfish or not
    :type isredfish: bool
    """

    def __init__(self, restresp=None, isredfish=True):
        self._resp = restresp
        self._patches = list()
        # Check if typedef can be used here
        self._typestring = "@odata.type" if isredfish else "Type"
        self.modified = False
        self.defpath = self.deftype = self.defetag = self._type = None
        self.__bool__ = self.__nonzero__

    @property
    def type(self):
        """Get type of the monolith member's response"""
        try:
            if self and self._typestring in self.resp.dict:
                return self.resp.dict[self._typestring]
            # Added for object type
            elif self and "type" in self.resp.dict:
                return self.resp.dict["type"]
        except (AttributeError, ValueError, TypeError):
            return self.deftype  # data not yet fetched, probably empty dict, so assume deftype
        return None

    @property
    def maj_type(self):
        """Get major type of the monolith member's response"""
        if self.type:
            if "." in self.type:
                types = ".".join(self.type.split(".", 2)[:2])
                retval = types[1:] if types.startswith("#") else types
            else:
                retval = self.type
            return retval
        return self.deftype

    def __nonzero__(self):
        """Defining the bool value for the class"""
        return True if self.resp else False

    @property
    def resp(self):
        """Get the entire response of the monolith member"""
        return self._resp

    @property
    def path(self):
        """Get path of the monolith member's response"""

        try:
            if self:
                return self.resp.request.path
        except (AttributeError, ValueError):
            pass
        return self.defpath

    @property
    def patches(self):
        """Get patches for the monolith member"""
        return self._patches

    @patches.setter
    def patches(self, val):
        """Set patches for the monolith member"""
        self._patches = val

    @property
    def dict(self):
        """Get the dictionary of the monolith member's response"""
        return self._resp.dict

    @property
    def etag(self):
        """Get the etag of the response"""
        return self.defetag if not self.resp else self.resp.getheader("etag")

    def popdefs(self, typename, pathval, etagval):
        """Populate the default values in the class

        :param typename: The default **Type** string. Example: @odata.type
        :type typename: str
        :param pathval: The default **Path** string. Example: @odata.id
        :type pathval: str
        :param etagval: The default **ETag** value.
        :type etagval: str
        """
        self.defetag = etagval
        self.deftype = typename
        self.defpath = pathval

    def to_dict(self):
        """Converts Monolith Member to a dictionary. This is the reverse of :func:`load_from_dict`.

        :returns: returns the Monolith Member in dictionary form
        """
        result = OrderedDict()
        if self.maj_type:
            result["Type"] = self.type

            if self.resp:
                if self.maj_type == "Collection.1" and "MemberType" in self.resp.dict:
                    result["MemberType"] = self.resp.dict["MemberType"]

            result["links"] = OrderedDict()
            result["links"]["href"] = ""
            result["ETag"] = self.etag

            if self.resp:
                result["Content"] = self.resp.dict
                result["Status"] = self.resp.status
                result["Headers"] = self.resp.getheaders()
            result["OriginalUri"] = self.path
            result["Patches"] = self._patches
            result["modified"] = self.modified
            result["MajType"] = self.maj_type

        return result

    def load_from_dict(self, src):
        """Load variables to a monolith member from a dictionary.
        This is the reverse of :func:`to_dict`.

        :param src: Source to load member data from.
        :type src: dict
        """
        if "Type" in src:
            self._type = src["Type"]
            if "Content" in src:
                restreq = RestRequest(method="GET", path=src["OriginalUri"])
                src["restreq"] = restreq
                self._resp = StaticRestResponse(**src)
            self.deftype = src["MajType"]
            self.defpath = src["OriginalUri"]
            self.defetag = src["ETag"]
            self._patches = src["Patches"]
            self.modified = src["modified"]


class RisMonolith(Dictable):
    """Monolithic cache of RIS data. This takes a :class:`redfish.rest.v1.RestClient` and uses it to
    gather data from a server and saves it in a modifiable database called monolith.

    :param client: client to use for data retrieval.
    :type client: :class:`redfish.rest.v1.RestClient`
    :param typepath: The compatibility class to use for differentiating between Redfish/LegacyRest.
    :type typepath: :class:`redfish.rest.ris.Typesandpathdefines`
    :param directory_load: The flag to quick load using resource directory if possible.
           When set to True this will load paths, etags, and types, but not create responses for
           every monolith member. When responses are needed, they will need to be loaded separately.
    :type directory_load: bool
    """

    def __init__(self, client, typepath, directory_load=True):
        self._client = client
        self.name = "Monolithic output of RIS Service"
        self._visited_urls = list()
        self._type = None
        self._name = None
        self.progress = 0
        self._directory_load = directory_load
        self.is_redfish = self.client.is_redfish
        self.typesadded = defaultdict(set)
        self.paths = dict()
        self.ctree = defaultdict(set)
        self.colltypes = defaultdict(set)

        self.typepath = typepath
        self.collstr = self.typepath.defs.collectionstring
        self.etagstr = "ETag"
        if self.is_redfish:
            self._resourcedir = "/redfish/v1/ResourceDirectory/"
        else:
            self._resourcedir = "/rest/v1/ResourceDirectory"

        # MultiThreading
        self.get_queue = Queue()
        self.threads = []

    @property
    def directory_load(self):
        """The flag to gather information about a tree without downloading every path. Only usable
        on HPE systems with a ResourceDirectory. type"""
        return self._directory_load

    @directory_load.setter
    def directory_load(self, dir_load):
        """Set the directory_load flag"""
        self._directory_load = dir_load

    @property
    def type(self):
        """Return monolith version type"""
        return "Monolith.1.0.0"

    @property
    def visited_urls(self):
        """The urls visited by the monolith"""
        return list(set(self._visited_urls) | set(self.paths.keys()))

    @visited_urls.setter
    def visited_urls(self, visited_urls):
        """Set visited URLS."""
        self._visited_urls = visited_urls

    @property
    def types(self):
        """Returns list of types for members in the monolith

        :rtype: list
        """
        return list(self.typesadded.keys())

    @types.setter
    def types(self, member):
        """Adds a member to monolith

        :param member: Member created based on response.
        :type member: RisMonolithMemberv100
        """
        self.typesadded[member.maj_type].add(member.path)
        patches = []
        if member.path in list(self.paths.keys()):
            patches = self.paths[member.path].patches
        self.paths[member.path] = member
        self.paths[member.path].patches.extend([patch for patch in patches])

    @property
    def client(self):
        """Returns the current client object reference

        :rtype: class object
        """
        return self._client

    @client.setter
    def client(self, curr_client):
        """Set the current client

        :param curr_client: current client object
        :type curr_client: class object
        """
        self._client = curr_client

    def path(self, path):
        """Provides the member corresponding to the path specified. Case sensitive.

        :param path: path of the monolith member to return
        :type path: str
        :rtype: RisMonolithMemberv100
        """
        try:
            return self.paths[path]
        except:
            return None

    def iter(self, typeval=None):
        """An iterator that yields each member of monolith associated with a specific type. In the
        case that no type is included this will yield all members in the monolith.

        :rtype: RisMonolithMemberv100
        """
        if not typeval:
            for _, val in list(self.paths.items()):
                yield val
        else:
            for typename in self.gettypename(typeval):
                for item in self.typesadded[typename]:
                    yield self.paths[item]

    #             types = next(self.gettypename(typeval), None)
    #             if types in self.typesadded:
    #                 for item in self.typesadded[types]:
    #                     yield self.paths[item]
    #             else:
    #                 raise RisInstanceNotFoundError("Unable to locate instance for" \
    #                                                             " '%s'\n" % typeval)

    def itertype(self, typeval):
        """Iterator that yields member(s) of given type in the monolith and raises an error if no
        member of that type is found.

        :param typeval: type name of the requested member.
        :type typeval: str
        :rtype: RisMonolithMemberv100
        """
        typeiter = self.gettypename(typeval)
        types = next(typeiter, None)
        if types:
            while types:
                for item in self.typesadded[types]:
                    yield self.paths[item]
                types = next(typeiter, None)
        else:
            raise RisInstanceNotFoundError("Unable to locate instance for '%s'\n" % typeval)

    def typecheck(self, types):
        """Check if a member of given type exists in the monolith

        :param types: type to check.
        :type types: str
        :rtype: bool
        """
        if any(types in val for val in self.types):
            return True
        return False

    def gettypename(self, types):
        """Takes a full type response and returns all major types associated.
        Example: #Type.v1_0_0.Type will return iter(Type.1)

        :param types: The type of the requested response.
        :type types: str
        :rtype: iter of major types
        """
        types = types[1:] if types[0] in ("#", "#") else types
        return iter((xt for xt in self.types if xt and types.lower() in xt.lower()))

    def update_member(self, member=None, resp=None, path=None, init=True):
        """Adds member to the monolith. If the member already exists the
        data is updated in place. Takes either a RisMonolithMemberv100 instance or a
        :class:`redfish.rest.containers.RestResponse` along with that responses path.

        :param member: The monolith member to add to the monolith.
        :type member: RisMonolithMemberv100
        :param resp: The rest response to add to the monolith.
        :type resp: :class:`redfish.rest.containers.RestResponse`
        :param path: The path correlating to the response.
        :type path: str
        :param init: Flag if addition is part of the initial load. Set this to false if you are
                     calling this by itself.
        :type init: bool
        """
        if not member and resp and path:
            self._visited_urls.append(path.lower())

            member = RisMonolithMemberv100(resp, self.is_redfish)
            if not member:  # Assuming for lack of member and not member.type
                return
            if not member.type:
                member.deftype = "object"  # Hack for general schema with no type

        self.types = member

        if init:
            self.progress += 1
            if LOGGER.getEffectiveLevel() == 40:
                self._update_progress()

    def load(
        self,
        path=None,
        includelogs=False,
        init=False,
        crawl=True,
        loadtype="href",
        loadcomplete=False,
        path_refresh=False,
        json_out=False,
    ):
        """Walks the entire data model and caches all responses or loads an individual path into
        the monolith. Supports both threaded and sequential crawling.

        :param path: The path to start the crawl from the provided path if crawling or
                     loads the path into monolith. If path is not included, crawl will start with
                     the default. The default is */redfish/v1/* or */rest/v1* depending on if the
                     system is Redfish or LegacyRest.
        :type path: str.
        :param includelogs: Flag to determine if logs should be downloaded as well in the crawl.
        :type includelogs: bool
        :param init: Flag to determine if this is the initial load.
        :type init: bool
        :param crawl: Flag to determine if load should crawl through found links.
        :type crawl: bool
        :param loadtype: Flag to determine if loading standard links: *href* or schema links: *ref*.
        :type loadtype: str.
        :param loadcomplete: Flag to download the entire data model including registries and
                             schemas.
        :type loadcomplete: bool
        :param path_refresh: Flag to reload the path specified, clearing any patches and overwriting the
                    current data in the monolith.
        :type path_refresh: bool
        """
        if init:
            if LOGGER.getEffectiveLevel() == 40 and not json_out:
                sys.stdout.write("Discovering data...")
            else:
                LOGGER.info("Discovering data...")
            self.name = self.name + " at %s" % self.client.base_url

        selectivepath = path
        if not selectivepath:
            selectivepath = self.client.default_prefix
        if loadtype == "href" and not self.client.base_url.startswith("blobstore://."):
            if not self.threads:
                for _ in range(6):
                    workhand = LoadWorker(self.get_queue)
                    workhand.setDaemon(True)
                    workhand.start()
                    self.threads.append(workhand)

            self.get_queue.put(
                (
                    selectivepath,
                    includelogs,
                    loadcomplete,
                    crawl,
                    path_refresh,
                    init,
                    None,
                    None,
                    self,
                )
            )
            self.get_queue.join()

            # Raise any errors from threads, and set them back to None after
            excp = None
            for thread in self.threads:
                if excp is None:
                    excp = thread.get_exception()
                thread.exception = None

            if excp:
                raise excp

            # self.member_queue.join()
        else:
            # We can't load ref or local client in a threaded manner
            self._load(
                selectivepath,
                originaluri=None,
                crawl=crawl,
                includelogs=includelogs,
                init=init,
                loadtype=loadtype,
                loadcomplete=loadcomplete,
                path_refresh=path_refresh,
                prevpath=None,
            )

        if init:
            if LOGGER.getEffectiveLevel() == 40 and not json_out:
                sys.stdout.write("Done\n")
            else:
                LOGGER.info("Done\n")
        if self.directory_load and init:
            self._populatecollections()

    def _load(
        self,
        path,
        crawl=True,
        originaluri=None,
        includelogs=False,
        init=True,
        loadtype="href",
        loadcomplete=False,
        path_refresh=False,
        prevpath=None,
    ):
        """Sequential version of loading monolith and parsing schemas.

        :param path: path to start load from.
        :type path: str
        :param crawl: flag to determine if load should traverse found links.
        :type crawl: bool
        :param originaluri: variable to assist in determining originating path.
        :type originaluri: str
        :param includelogs: flag to determine if logs should be downloaded also.
        :type includelogs: bool
        :param init: flag to determine if first run of load.
        :type init: bool
        :param loadtype: flag to determine if load is meant for only href items.
        :type loadtype: str.
        :param loadcomplete: flag to download the entire monolith
        :type loadcomplete: bool
        :param path_refresh: flag to reload the members in the monolith instead of skip if they exist.
        :type path_refresh: bool
        """

        if (path.endswith("?page=1") or path.endswith(".json")) and not loadcomplete:
            # Don't download schemas in crawl unless we are loading absolutely everything
            return
        elif not includelogs and crawl:
            # Only include logs when asked as there can be an extreme amount of entries
            if "/log" in path.lower():
                return

        # remove fragments and quote characters in path
        p = urlparse(path)
        p = ParseResult(
            scheme=p.scheme, netloc=p.netloc, path=quote(p.path), params=p.params, query=p.query, fragment=""
        )
        path = urlunparse(p)

        if prevpath and prevpath != path:
            self.ctree[prevpath].update([path])
        if not path_refresh:
            if path.lower() in self.visited_urls:
                return
        LOGGER.debug("_loading %s", path)

        resp = self.client.get(path)
        if resp.status != 200 and path.lower() == self.typepath.defs.biospath:
            raise BiosUnregisteredError()
        elif resp.status == 401:
            raise SessionExpired("Invalid session. Please logout and " "log back in or include credentials.")
        elif resp.status not in (201, 200):
            self.removepath(path)
            return

        if loadtype == "ref":
            try:
                if resp.status in (201, 200):
                    self.update_member(resp=resp, path=path, init=init)
                self._parse_schema(resp)
            except jsonpointer.JsonPointerException:
                raise SchemaValidationError()

        self.update_member(resp=resp, path=path, init=init)

        fpath = lambda pa, path: (
            path if pa.endswith(self.typepath.defs.hrefstring) and pa.startswith((self.collstr, "Entries")) else None
        )

        if loadtype == "href":
            # follow all the href attributes
            if self.is_redfish:
                jsonpath_expr = jsonpath_rw.parse("$..'@odata.id'")
            else:
                jsonpath_expr = jsonpath_rw.parse("$..href")
            matches = jsonpath_expr.find(resp.dict)

            if "links" in resp.dict and "NextPage" in resp.dict["links"]:
                if originaluri:
                    next_link_uri = originaluri + "?page=" + str(resp.dict["links"]["NextPage"]["page"])
                    href = "%s" % next_link_uri

                    self._load(
                        href,
                        originaluri=originaluri,
                        includelogs=includelogs,
                        crawl=crawl,
                        init=init,
                        prevpath=None,
                        loadcomplete=loadcomplete,
                    )
                else:
                    next_link_uri = path + "?page=" + str(resp.dict["links"]["NextPage"]["page"])

                    href = "%s" % next_link_uri
                    self._load(
                        href,
                        originaluri=path,
                        includelogs=includelogs,
                        crawl=crawl,
                        init=init,
                        prevpath=None,
                        loadcomplete=loadcomplete,
                    )

            # Only use monolith if we are set to
            matchrdirpath = (
                next((match for match in matches if match.value == self._resourcedir), None)
                if self.directory_load
                else None
            )
            if not matchrdirpath and crawl:
                for match in matches:
                    if path == "/rest/v1" and not loadcomplete:
                        if (
                            str(match.full_path) == "links.Schemas.href"
                            or str(match.full_path) == "links.Registries.href"
                        ):
                            continue
                    elif not loadcomplete:
                        if (
                            str(match.full_path) == "Registries.@odata.id"
                            or str(match.full_path) == "JsonSchemas.@odata.id"
                        ):
                            continue

                    if match.value == path:
                        continue
                    elif not isinstance(match.value, six.string_types):
                        continue

                    href = "%s" % match.value
                    self._load(
                        href,
                        crawl=crawl,
                        originaluri=originaluri,
                        includelogs=includelogs,
                        init=init,
                        prevpath=fpath(str(match.full_path), path),
                        loadcomplete=loadcomplete,
                    )
            elif crawl:
                href = "%s" % matchrdirpath.value
                self._load(
                    href,
                    crawl=crawl,
                    originaluri=originaluri,
                    includelogs=includelogs,
                    init=init,
                    prevpath=path,
                    loadcomplete=loadcomplete,
                )
            if loadcomplete:
                if path == "/rest/v1":
                    schemamatch = jsonpath_rw.parse("$..extref")
                else:
                    schemamatch = jsonpath_rw.parse("$..Uri")
                smatches = schemamatch.find(resp.dict)
                matches = matches + smatches
                for match in matches:
                    if isinstance(match.value, six.string_types):
                        self._load(
                            match.value,
                            crawl=crawl,
                            originaluri=originaluri,
                            includelogs=includelogs,
                            init=init,
                            loadcomplete=loadcomplete,
                            prevpath=fpath(str(match.full_path), path),
                        )

    def _parse_schema(self, resp):
        """Function to get and replace schema $ref with data

        :param resp: response data containing ref items.
        :type resp: str
        """
        # pylint: disable=maybe-no-member
        if not self.typepath.gencompany:
            return self._parse_schema_gen(resp)
        jsonpath_expr = jsonpath_rw.parse('$.."$ref"')
        matches = jsonpath_expr.find(resp.dict)
        respcopy = resp.dict
        typeregex = "([#,@].*?\.)"
        if matches:
            for match in matches:
                fullpath = str(match.full_path)
                jsonfile = match.value.split("#")[0]
                jsonpath = match.value.split("#")[1]
                listmatch = None
                found = None

                if "redfish.dmtf.org" in jsonfile:
                    if "odata" in jsonfile:
                        jsonpath = jsonpath.replace(jsonpath.split("/")[-1], "odata" + jsonpath.split("/")[-1])
                    jsonfile = "Resource.json"

                found = re.search(typeregex, fullpath)
                if found:
                    repitem = fullpath[found.regs[0][0] : found.regs[0][1]]
                    schemapath = "/" + fullpath.replace(repitem, "~").replace(".", "/").replace("~", repitem)
                else:
                    schemapath = "/" + fullpath.replace(".", "/")

                if ".json" in jsonfile:
                    itempath = schemapath

                    if self.is_redfish:
                        if resp.request.path[-1] == "/":
                            newpath = "/".join(resp.request.path.split("/")[:-2]) + "/" + jsonfile + "/"
                        else:
                            newpath = "/".join(resp.request.path.split("/")[:-1]) + "/" + jsonfile + "/"
                    else:
                        newpath = "/".join(resp.request.path.split("/")[:-1]) + "/" + jsonfile

                    if "href.json" in newpath:
                        continue

                    if newpath.lower() not in self.visited_urls:
                        self.load(
                            newpath,
                            crawl=False,
                            includelogs=False,
                            init=False,
                            loadtype="ref",
                        )

                    instance = list()

                    # deprecated type "string" for Type.json
                    if "string" in self.types:
                        for item in self.iter("string"):
                            instance.append(item)
                    if "object" in self.types:
                        for item in self.iter("object"):
                            instance.append(item)

                    for item in instance:
                        if jsonfile in item.path:
                            if "anyOf" in fullpath:
                                break

                            dictcopy = item.dict
                            try:
                                # TODO may need to really verify this is acceptable regex
                                listmatch = re.search("[][0-9]+[]", itempath)
                            except regexerr:
                                pass
                                # LOGGER.info("An error occurred with regex match on path: %s\n%s\n"\
                                #            % (itempath, str(excp)))

                            if listmatch:
                                start = listmatch.regs[0][0]
                                end = listmatch.regs[0][1]

                                newitempath = [itempath[:start], itempath[end:]]
                                start = jsonpointer.JsonPointer(newitempath[0])
                                end = jsonpointer.JsonPointer(newitempath[1])

                                del start.parts[-1], end.parts[-1]
                                vals = start.resolve(respcopy)

                                count = 0

                                for val in vals:
                                    try:
                                        if "$ref" in six.iterkeys(end.resolve(val)):
                                            end.resolve(val).pop("$ref")
                                            end.resolve(val).update(dictcopy)
                                            replace_pointer = jsonpointer.JsonPointer(end.path + jsonpath)

                                            data = replace_pointer.resolve(val)
                                            set_pointer(val, end.path, data)
                                            start.resolve(respcopy)[count].update(val)

                                            break
                                    except:
                                        count += 1
                            else:
                                itempath = jsonpointer.JsonPointer(itempath)
                                del itempath.parts[-1]

                                try:
                                    if "$ref" in six.iterkeys(itempath.resolve(respcopy)):
                                        itempath.resolve(respcopy).pop("$ref")
                                        itempath.resolve(respcopy).update(dictcopy)
                                        break
                                except jsonpointer.JsonPointerException:
                                    pass

                if jsonpath:
                    if "anyOf" in fullpath:
                        continue

                    if not jsonfile:
                        replacepath = jsonpointer.JsonPointer(jsonpath)
                        schemapath = schemapath.replace("/$ref", "")
                        if re.search("\[\d]", schemapath):
                            schemapath = schemapath.translate(str.maketrans("", "", "[]"))
                        schemapath = jsonpointer.JsonPointer(schemapath)
                        data = replacepath.resolve(respcopy)

                        if "$ref" in schemapath.resolve(respcopy):
                            schemapath.resolve(respcopy).pop("$ref")
                            schemapath.resolve(respcopy).update(data)

                    else:
                        if not listmatch:
                            schemapath = schemapath.replace("/$ref", "")
                            replacepath = schemapath + jsonpath
                            replace_pointer = jsonpointer.JsonPointer(replacepath)
                            try:
                                data = replace_pointer.resolve(respcopy)
                                set_pointer(respcopy, schemapath, data)
                            except jsonpointer.JsonPointerException:
                                # TODO
                                pass

            resp.loaddict(respcopy)
        else:
            resp.loaddict(respcopy)

    def _parse_schema_gen(self, resp):
        """Redfish general function to get and replace schema $ref with data

        :param resp: response data containing ref items.
        :type resp: str

        """
        # pylint: disable=maybe-no-member
        getval = lambda inmat: getval(inmat.left) + "/" + str(inmat.right) if hasattr(inmat, "left") else str(inmat)
        respcopy = resp.dict
        jsonpath_expr = jsonpath_rw.parse('$.."anyOf"')
        while True:
            matches = jsonpath_expr.find(respcopy)
            if not matches:
                break
            match = matches[0]
            newval = None
            schlist = match.value
            schlist = [ele for ele in list(schlist) if ele != {"type": "null"}]
            norefsch = [ele for ele in list(schlist) if isinstance(ele, dict) and len(ele.keys()) > 1]
            if norefsch:
                newval = norefsch[0]
            else:
                newsc = [ele for ele in list(schlist) if not ele["$ref"].split("#")[0]]
                newval = newsc[0] if newsc else None
                if not newval:
                    schlist = [
                        ele["$ref"]
                        for ele in list(schlist)
                        if "$ref" in ele.keys()
                        and (ele["$ref"].split("#")[0].endswith(".json") and "odata" not in ele["$ref"].split("#")[0])
                    ]
                    maxsch = max(schlist)
                    newval = {"$ref": maxsch}

            itempath = "/" + getval(match.full_path)
            if re.search("\[\d+]", itempath):
                itempath = itempath.translate(str.maketrans("", "", "[]"))
            itempath = jsonpointer.JsonPointer(itempath)
            del itempath.parts[-1]
            if "anyOf" in six.iterkeys(itempath.resolve(respcopy)):
                itempath.resolve(respcopy).pop("anyOf")
                itempath.resolve(respcopy).update(newval)

        jsonpath_expr = jsonpath_rw.parse('$.."$ref"')
        matches = jsonpath_expr.find(respcopy)
        if matches:
            for _, match in enumerate(matches):
                jsonfile = match.value.split("#")[0]
                jsonfile = "" if jsonfile.lower() == resp.request.path.lower() else jsonfile
                jsonpath = match.value.split("#")[1]

                schemapath = "/" + getval(match.full_path)
                if jsonfile:
                    itempath = schemapath
                    if "/" not in jsonfile:
                        inds = -2 if resp.request.path[-1] == "/" else -1
                        jsonfile = "/".join(resp.request.path.split("/")[:inds]) + "/" + jsonfile + "/"
                    if jsonfile not in self.paths:
                        self.load(
                            jsonfile,
                            crawl=False,
                            includelogs=False,
                            init=False,
                            loadtype="ref",
                        )
                    item = self.paths[jsonfile] if jsonfile in self.paths else None

                    if not item:
                        if "anyOf" not in schemapath:
                            raise SchemaValidationError()
                        continue
                    if re.search("\[\d+]", itempath):
                        itempath = itempath.translate(str.maketrans("", "", "[]"))
                    itempath = jsonpointer.JsonPointer(itempath)
                    del itempath.parts[-1]
                    try:
                        if "$ref" in six.iterkeys(itempath.resolve(respcopy)):
                            itempath.resolve(respcopy).pop("$ref")
                            itempath.resolve(respcopy).update(item.dict)
                    except jsonpointer.JsonPointerException:
                        pass

                if jsonpath:
                    schemapath = schemapath.replace("/$ref", "")
                    if re.search("\[\d+]", schemapath):
                        schemapath = schemapath.translate(str.maketrans("", "", "[]"))
                    if not jsonfile:
                        replacepath = jsonpointer.JsonPointer(jsonpath)
                        schemapath = jsonpointer.JsonPointer(schemapath)
                        data = replacepath.resolve(respcopy)
                        if "$ref" in schemapath.resolve(respcopy):
                            schemapath.resolve(respcopy).pop("$ref")
                            schemapath.resolve(respcopy).update(data)
                    else:
                        replacepath = schemapath + jsonpath
                        replace_pointer = jsonpointer.JsonPointer(replacepath)
                        data = replace_pointer.resolve(respcopy)
                        set_pointer(respcopy, schemapath, data)

            resp.loaddict(respcopy)
        else:
            resp.loaddict(respcopy)

    def load_from_dict(self, src):
        """Load data to monolith from a dict. This is the reverse of :func:`to_dict`.

        :param src: data receive from rest operation.
        :type src: str

        """
        self._type = src["Type"]
        self._name = src["Name"]
        self.typesadded = defaultdict(set, {ki: set(val) for ki, val in src["typepath"].items()})
        self.ctree = defaultdict(set, {ki: set(val) for ki, val in src["ctree"].items()})
        self.colltypes = defaultdict(set, {ki: set(val) for ki, val in src["colls"].items()})
        for _, resp in list(src["resps"].items()):
            member = RisMonolithMemberv100(None, self.is_redfish)
            member.load_from_dict(resp)
            self.update_member(member=member, init=False)

    def to_dict(self):
        """Convert data to a dict from monolith. This is the reverse of :func:`load_from_dict`."""
        result = OrderedDict()
        result["Type"] = self.type
        result["Name"] = self.name
        result["typepath"] = self.typesadded
        result["ctree"] = self.ctree
        result["colls"] = self.colltypes
        result["resps"] = {x: v.to_dict() for x, v in list(self.paths.items())}
        return result

    def markmodified(self, opath, path=None, modpaths=None):
        """Mark the paths to be modifed which are connected to current path. When calling this
        function you only need to include `opath`.

        :param opath: original path which has been modified
        :type opath: str
        """
        modpaths = set() if modpaths is None else modpaths
        path = path if path else opath
        if not path:
            return
        modpaths.update(self.ctree[path] if path in self.ctree else set())
        self.paths[path].modified = True
        for npath in [
            unmodpath for unmodpath in modpaths if unmodpath in self.paths and not self.paths[unmodpath].modified
        ]:
            self.markmodified(opath, path=npath, modpaths=modpaths)
        return modpaths

    def checkmodified(self, opath, path=None, modpaths=None):
        """Check if the path or its children are modified. When calling this
        function you only need to include `opath`.

        :param opath: original path which has been modified
        :type opath: str
        """
        # return [paths for paths in self.ctree[path] if self.paths[paths].modified]
        modpaths = set() if modpaths is None else modpaths
        path = path if path else opath
        newpaths = set()
        if not path:
            return
        if path in self.paths and self.paths[path].modified:
            newpaths = (
                set([conn for conn in self.ctree[path] if conn in self.paths and self.paths[path].modified]) - modpaths
            )
            modpaths.update(newpaths | set([path]))
        for npath in [unmodpath for unmodpath in newpaths]:
            self.checkmodified(opath, path=npath, modpaths=modpaths)
        return modpaths

    def removepath(self, path):
        """Remove a given path from the cache

        :param path: path which is to be checked if modified
        :type path: str
        """
        if path in self._visited_urls:
            self._visited_urls.remove(path)
        if path not in self.paths:
            return
        if path in self.typesadded[self.paths[path].maj_type]:
            self.typesadded[self.paths[path].maj_type].remove(path)
        if not self.typesadded[self.paths[path].maj_type]:
            del self.typesadded[self.paths[path].maj_type]
        del self.paths[path]
        if path in self.ctree:
            del self.ctree[path]
        _ = [self.ctree[paths].remove(path) for paths in self.ctree if path in self.ctree[paths]]

    def _populatecollections(self):
        """Populate the collections type and types depending on resourcedirectory"""
        if self._resourcedir not in self.paths:
            return
        self.colltypes = defaultdict(set)
        alltypes = []
        colls = []
        for item in self.paths[self._resourcedir].dict["Instances"]:
            # Fix for incorrect RDir instances.
            if self.typepath.defs.typestring not in item or item[self.typepath.defs.hrefstring] in self.paths:
                continue
            typename = ".".join(item[self.typepath.defs.typestring].split(".", 2)[:2]).split("#")[-1]
            _ = [alltypes.append(typename) if "Collection" not in typename else None]
            _ = [colls.append(typename) if "Collection" in typename else None]
            member = RisMonolithMemberv100(None, self.is_redfish)
            member.popdefs(typename, item[self.typepath.defs.hrefstring], item[self.etagstr])
            self.update_member(member=member, init=False)
        for coll in colls:
            collname = coll.split("Collection")[0].split("#")[-1]
            typename = next((name for name in alltypes if name.startswith(collname)), None)
            colltype = ".".join(coll.split(".", 2)[:2]).split("#")[-1]
            self.colltypes[typename].add(colltype)

    def capture(self, redmono=False):
        """Crawls the server specified by the client and returns the entire monolith.

        :param redmono: Flag to return only the headers and responses instead of the entire monolith
                        member data.
        :type redmono: bool
        :rtype: dict
        """
        self.load(includelogs=True, crawl=True, loadcomplete=True, path_refresh=True, init=True)
        return (
            self.to_dict()
            if not redmono
            else {
                x: {"Headers": v.resp.getheaders(), "Response": v.resp.dict} for x, v in list(self.paths.items()) if v
            }
        )

    def killthreads(self):
        """Function to kill threads on logout"""
        threads = []
        for thread in threading.enumerate():
            if isinstance(thread, LoadWorker):
                self.get_queue.put(
                    (
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                        "KILL",
                    )
                )
                threads.append(thread)

        for thread in threads:
            thread.join()

    def _update_progress(self):
        """Simple function to increment the dot progress"""
        if self.progress % 6 == 0:
            sys.stdout.write(".")



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Shared types used in this library."""

# ---------Imports---------

import logging

import jsonpatch

from redfish.rest.containers import JSONEncoder

# ---------End of imports---------

# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)

# ---------End of debug logger---------


class JSONEncoder(JSONEncoder):
    """Custom JSONEncoder that understands our types"""

    def default(self, obj):
        """Set defaults

                :param obj: json object.
        :type obj: str.

        """
        if isinstance(obj, Dictable):
            return obj.to_dict()
        elif isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, jsonpatch.JsonPatch):
            return obj.patch
        return super(JSONEncoder, self).default(obj)


class Dictable(object):
    """A base class which adds the to_dict method used during json encoding"""

    def to_dict(self):
        """Overridable funciton"""
        raise RuntimeError("You must override this method in your derived class")



""" Redfish restful library """

__all__ = ["rest", "ris", "hpilo"]
__version__ = "5.3.0.0"

import logging

from redfish.rest.v1 import AuthMethod, LegacyRestClient, RedfishClient


def redfish_logger(file_name, log_format, log_level=logging.ERROR):
    """redfish logger"""
    formatter = logging.Formatter(log_format)
    fhdl = logging.FileHandler(file_name)
    fhdl.setFormatter(formatter)
    logger = logging.getLogger(__name__)
    logger.addHandler(fhdl)
    logger.setLevel(log_level)
    return logger



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of modifying a user account
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def modify_ilo_user_account(_redfishobj, username_to_modify, new_loginname, new_username, \
                         new_password, role_id, privilege_dict):

    account_collection_uri = None

    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually to obtain
        #account info
        account_service_uri = _redfishobj.root.obj['AccountService']['@odata.id']
        account_service_response = _redfishobj.get(account_service_uri)
        account_collection_uri = account_service_response.obj['Accounts']['@odata.id']
        #modify role id
        if role_id:
            body = {"RoleId": role_id}
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#ManagerAccountCollection.' in instance['@odata.type']:
                account_collection_uri = instance['@odata.id']
        if privilege_dict:
            #HPE server, so modify privileges
            body = {"Oem": {"Hpe": {"Privileges": {}}}}
            for priv in privilege_dict:
                body["Oem"]["Hpe"]["Privileges"][priv] = privilege_dict[priv]
        if new_loginname:
            #modify login name
            body["Oem"]["Hpe"]["LoginName"] = new_loginname

    if new_username:
        body["UserName"] = new_username
    if new_password:
        body["Password"] = new_password

    #find the account to modify
    account_uri_to_modify = None
    account_uris = REDFISHOBJ.get(account_collection_uri)
    for account_uri in account_uris.dict['Members']:
        account = REDFISHOBJ.get(account_uri['@odata.id'])
        if account.dict['UserName'] == username_to_modify:
            account_uri_to_modify = account_uri['@odata.id']
            break

    if not account_uri_to_modify:
        sys.stderr.write("Cannot find account to modify")
        return

    #modify the account
    resp = REDFISHOBJ.patch(account_uri_to_modify, body)

    #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
    #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Success!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #username of the account to modify
    USERNAME_TO_MODIFY = "bruce_wayne"

    #account login name to change the account to
    NEW_LOGINNAME = "joker"

    #account user name to change the account to
    NEW_USERNAME = "joker"

    #account password to change the account to
    NEW_PASSWORD = "joker123"

    #role to change account to
    ROLE_ID = "ReadOnly" #Administrator, ReadOnly or Operator are available

    #update HPE account privileges
    PRIVILEGE_DICT = {"iLOConfigPriv": False, "VirtualMediaPriv": False, "RemoteConsolePriv": True,\
                      "UserConfigPriv": False, "VirtualPowerAndResetPriv": False, \
                      "SystemRecoveryConfigPriv": False, "LoginPriv": True, \
                      "HostStorageConfigPriv": False, "HostNICConfigPriv": False, \
                      "HostBIOSConfigPriv": False}
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    modify_ilo_user_account(REDFISHOBJ, USERNAME_TO_MODIFY, NEW_LOGINNAME, NEW_USERNAME, \
                            NEW_PASSWORD, ROLE_ID, PRIVILEGE_DICT)

    REDFISHOBJ.logout()
